###############################################################################
##### Module with emgfit models for Gaussian and Hyper-EMG distributions
##### Author: Stefan Paul

##### Import dependencies
import numpy as np
import warnings
from .config import *
from .emg_funcs import *
from .model import EMGModel
from scipy.special import erfcx
from numpy import sqrt, pi

    
###### Define parameter bounds
# Variables ending with _nsigma define bounds in multiples of the initial value
# for the sigma parameter of the underlying Gaussian ('<initial_sigma>')
MU_VAR_NSIGMA = 5 # bound mus to <mu0> +- mu_var_sigma*<initial_sigma>
SIGMA_MIN = 1e-15 # lower bound of sigmas
SIGMA_MAX_NSIGMA = 10 # upper bound of sigmas = SIGMA_MAX_NSIGMA*<initial_sigma>
TAU_MIN = 1e-15 # lower bound of taus
TAU_MAX_NSIGMA = 500 # upper bound of taus = TAU_MAX_NSIGMA*<initial_sigma>
AMP_MIN = 1e-20

def create_default_init_pars(mass_number=100, resolving_power=3e05): 
    """
    Scale default parameters to mass of interest and return parameter dictionary.

    Parameters
    ----------
    mass_number : int, optional
        Atomic mass number of peaks of interest, defaults to 100.
    resolving_power : float, optional
        Typical resolving power of the spectrometer  at FWHM level. 
        Defaults to 3e05.

    Returns
    -------
    dict
        Dictionary with default initial parameters (scaled to `mass_number`).

    Notes
    -----
    **The default parameters were defined for mass 100**, to obtain suitable
    parameters at other masses all mass-dependent parameters (i.e. shape
    parameters & `amp`) are multiplied by the scaling factor `mass_number`/100.

    The standard deviation of the underlying Gaussian :math:`\sigma` is 
    calculated as :math::`\\sigma = A / (R 2 \\sqrt(2 \\ln(2))`, where 
    :math:`A` denotes the specified `mass_number` and :math:`R` is the given 
    `resolving_power`.  

    """
    # Default initial parameters for peaks around mass 100 (with
    # mass scaling factor):
    scl_factor = mass_number/100
    sigma_to_FWHM = 2*np.sqrt(2*np.log(2))  # for Gaussian peak
    amp = 0.45*scl_factor
    mu = None  # flag for below that this is a generic shape parameter set
    sigma = mass_number/(resolving_power*sigma_to_FWHM) # [u]
    theta = 0.5
    eta_m1 = 0.85
    eta_m2 = 0.10
    eta_m3 = 0.05
    tau_m1 =   50e-06*scl_factor # [u]
    tau_m2 =  500e-06*scl_factor # [u]
    tau_m3 = 1000e-06*scl_factor # [u]
    eta_p1 = 0.85
    eta_p2 = 0.10
    eta_p3 = 0.05
    tau_p1 =   50e-06*scl_factor # [u]
    tau_p2 =  500e-06*scl_factor # [u]
    tau_p3 = 1000e-06*scl_factor # [u]
    pars_dict = {'amp': amp, 'mu': mu, 'sigma': sigma, 'theta': theta,
                 'eta_m1': eta_m1, 'eta_m2': eta_m2, 'eta_m3': eta_m3,
                 'tau_m1': tau_m1, 'tau_m2': tau_m2, 'tau_m3': tau_m3,
                 'eta_p1': eta_p1, 'eta_p2': eta_p2, 'eta_p3': eta_p3,
                 'tau_p1': tau_p1, 'tau_p2': tau_p2, 'tau_p3': tau_p3}
    return pars_dict

pars_dict = create_default_init_pars()


def scl_init_pars(init_pars, mu0=None, mu_ref=None, scl_coeff=1.0, decimals=9):
    """Scale initial shape parameters of reference peak to peak at `mu0`.

    If `scl_coeff` is None the original `init_pars` dictionary is returned.

    """
    from copy import deepcopy
    scaled_pars = deepcopy(init_pars)
    for pname, pval in scaled_pars.items():
        if pname.startswith(('sigma','tau')):
            mu_ratio = mu0/mu_ref if mu0 and mu_ref else 1
            scaled_pars[pname] *= scl_coeff*np.round(mu_ratio, decimals)
    return scaled_pars


def _enforce_shared_shape_pars(model, peak_index, index_ref_peak,
                               scale_shape_pars, scl_coeff, mu_ref):
    """Enfore shared peak-shapes and optionally scale the shape parameters

    Parameters
    ----------
    model : :class:`emgfit.model.EMGModel`
        Fit model to impose parameter constraints on.
    peak_index : int
        Index of the :class:`~emgfit.spectrum.peak` corresponding to `model`.
    index_ref_peak : int
        Index of the :class:`~emgfit.spectrum.peak` to be used as
        shape-reference peak.
    scale_shape_pars : bool
        Whether to scale the scale-dependent shape parameters adopted from the
        shape-reference peak.
    scl_coeff : float, optional
        Constant coefficient used to scale the scale-dependent shape parameters.
    mu_ref : float or str, optional
        Centroid of the underlying Gaussian of the shape-reference peak. This
        argument is only relevant when ``scale_shape_pars=True``. If `mu_ref` is
        set to a float, this number is used as the fixed (Gaussian) reference
        centroid for calculating the scale factor. If `mu_ref="varying"`,
        `mu_ref` is set to the varying (Gaussian) centroid shape-reference peak.

    Notes
    -----
    If `scale_shape_pars` is True, the model's scale-dependent shape parameters
    are multiplied with the scale factor ``scl_fac = scl_coeff``. If further
    `mu_ref` is not None, the scale-dependent shape parameters are multiplied
    with the scale factor ``scl_fac = mu/mu_ref * scl_coeff``, where `mu` is the
    centroid of the Gaussian underlying the specified `model`.  When the
    reference peak is among the peaks to be fitted, ``scl_fac`` will then be
    dynamically re-calculated from the `mu` and `mu_ref` values of a given
    iteration in a fit.

    """
    pref = 'p{0}_'.format(peak_index)
    ref_pref = 'p{0}_'.format(index_ref_peak)
    if scale_shape_pars is True:
        if mu_ref is None:
            model.set_param_hint(pref+'scl_fac', expr=str(scl_coeff))
        else:
            model.set_param_hint(pref+'scl_fac',
                                 expr=pref+'mu/'+ref_pref+'mu * '+str(scl_coeff))

        for ppname in model.param_names: # Set parameter constraints
            pname = model._strip_prefix(ppname)
            if pname.startswith(('theta','eta','delta')):
                model.set_param_hint(pref+pname, expr=ref_pref+pname)
            elif pname.startswith(('sigma','tau')):
                model.set_param_hint(pref+pname,
                                     expr=pref+'scl_fac * '+ref_pref+pname)
    else:
        for ppname in model.param_names: # Set parameter constraints
            pname = model._strip_prefix(ppname)
            if pname.startswith(('sigma','theta','eta','tau','delta')):
                model.set_param_hint(pref+pname, expr=ref_pref+pname)

    return model


def _calc_mu_emg(fit_model, pars, pref=""):
    """Calculate the mean of a Hyper-EMG peak from the best-fit parameters.

    Parameters
    ----------
    fit_model :
        Name of fit model used to obtain `pars`.
    pars : :class:`emgfit.parameter.Parameters`, optional
        Parameters object to obtain shape parameters for calculation from.
        This argument can be used when no fit result is available.
    pref : str, optional
        Prefix of peak parameters of interest.

    Returns
    -------
    float [u/z]
        Mean of Hyper-EMG fit of peak of interest.

    """
    if fit_model.startswith("emg"):
        N_left_tails = int(fit_model[3])
        N_right_tails = int(fit_model[4])
        li_eta_m, li_tau_m, li_eta_p, li_tau_p = [],[],[],[]
        for i in np.arange(1,N_left_tails+1):
            if N_left_tails == 1:
                li_eta_m = [1]
            else:
                li_eta_m.append(pars[pref+'eta_m'+str(i)].value)
            li_tau_m.append(pars[pref+'tau_m'+str(i)].value)
        for i in np.arange(1,N_right_tails+1):
            if N_right_tails == 1:
                li_eta_p = [1]
            else:
                li_eta_p.append(pars[pref+'eta_p'+str(i)].value)
            li_tau_p.append(pars[pref+'tau_p'+str(i)].value)
        if N_left_tails == 0:
            theta = 0
        elif N_right_tails == 0:
            theta = 1
        else:
            theta = pars[pref+'theta'].value
        mu_EMG = mu_emg(pars[pref+'mu'].value,
                        theta,
                        tuple(li_eta_m),tuple(li_tau_m),
                        tuple(li_eta_p),tuple(li_tau_p) )
    elif fit_model == "Gaussian":
        mu_EMG = pars[pref+'mu'].value
    else:
        raise Exception("`fit_result` is not from a Hyper-EMG fit.")

    return mu_EMG


def erfcxinv(y):
    """Approximate inverse of the scaled complementary error function

    This approximation is adapted from code by John D'Errico posted at
    https://www.mathworks.com/matlabcentral/answers/302915-inverse-of-erfcx-scaled-complementary-error-function

    """
    # erfcx inverse, for y no larger than 1.
    # for y <= 1, use the large x approximation for a starting value.
    mask = (y <= 1)
    ret = np.zeros_like(y)
    ret[mask] = 1./(y[mask]*sqrt(pi))
    # for y > 1, use exp(x^2) as a very rough approximation
    # to erfcx
    ret[~mask] = -sqrt(np.log(y[~mask]))
    for n in range(7):
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            denom = 2*ret*erfcx(ret) - 2/sqrt(pi)
            ret += np.where(denom==0, 0.0, - (erfcx(ret) - y)/denom)
    return ret


def get_mu0(x_m, init_pars, fit_model):
    """Estimate initial value of Gaussian centroid `mu` from the peak's mode

    Parameters
    ----------
    x_m : float
        Mode (i.e. x-position of the maximum) of the distribution.
    init_pars : dict
        Dictionary with initial values of the shape parameters.
    fit_model : str
        Name of used fit model (e.g. 'emg01', 'emg10', 'emg12'...).

    Notes
    -----
    For a Gaussian, the mean `mu` is simply equal to the `x_m` argument.

    For highly asymmetric hyper-EMG distributions the centroid of the underlying
    Gaussian `mu` can strongly deviate from the mode :math:`x_{m}` (i.e. the
    x-position of the peak maximum). Hence, the initial Gaussian centroid `mu`
    (:math:`\\mu`) is calculated by rearranging the equation for the mode of the
    hyper-EMG distribution:

    .. math::

      \\mu = x_{m}
             - \\theta\\sum_{i=1}^{N_-}\\eta_{-i}\\left(\\sqrt{2}\\sigma
               \\cdot\\mathrm{erfcxinv}\\left( \\frac{\\tau_{-i}}{\\sigma}
               \\sqrt{\\frac{2}{\\pi}}\\right) - \\frac{\\sigma^2}{\\tau_{-i}}
               \\right) \\\\
             + (1-\\theta)\\sum_{i=1}^{N_-}\\eta_{+i}\\left(\\sqrt{2}\\sigma
               \\cdot\\mathrm{erfcxinv}\\left( \\frac{\\tau_{+i}}{\\sigma}
               \\sqrt{\\frac{2}{\\pi}}\\right) - \\frac{\\sigma^2}{\\tau_{-i}}
               \\right),

    where the mode :math:`x_{m}` can be estimated by the peak marker position
    `x_pos` and :func:`emgfit.fit_models.erfcxinv` is the inverse of the scaled
    complementary error function.

    """
    if fit_model == 'Gaussian':
        mu0 = x_m
    elif fit_model.startswith('emg') and len(fit_model) == 5:
        sigma = init_pars['sigma']
        t_order_m = int(fit_model[-2])
        sum_m = 0
        if t_order_m == 1:
            tau_m1 = init_pars['tau_m1']
            sum_m = (sqrt(2)*sigma*erfcxinv(tau_m1/sigma*sqrt(2/pi))
                     - sigma**2/tau_m1)
        else:
            for i in range(t_order_m):
                eta_mi = init_pars['eta_m{}'.format(i+1)]
                tau_mi = init_pars['tau_m{}'.format(i+1)]
                sum_m += eta_mi*(sqrt(2)*sigma*erfcxinv(tau_mi/sigma*sqrt(2/pi))
                                 - sigma**2/tau_mi)

        t_order_p = int(fit_model[-1])
        sum_p = 0
        if t_order_p == 1:
            tau_p1 = init_pars['tau_p1']
            sum_p = (sqrt(2)*sigma*erfcxinv(tau_p1/sigma*sqrt(2/pi))
                     - sigma**2/tau_p1)
        else:
            for i in range(t_order_p):
                eta_pi = init_pars['eta_p{}'.format(i+1)]
                tau_pi = init_pars['tau_p{}'.format(i+1)]
                sum_p += eta_pi*(sqrt(2)*sigma*erfcxinv(tau_pi/sigma*sqrt(2/pi))
                                 - sigma**2/tau_pi)

        if t_order_p == 1 and t_order_m == 0:
            theta = 0
        elif t_order_p == 0 and t_order_m == 1:
            theta = 1
        else:
            theta = init_pars['theta']

        mu0 = x_m - theta*sum_m + (1-theta)*sum_p
    else:
        msg = str("fit_model must be 'Gaussian' or 'emgXY' with X & Y denoting the tail orders!")
        raise Exception(msg)
    
    return mu0


###############################################################################
##### Define emgfit fit models
def Gaussian(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Gaussian emgfit model (single-peak Gaussian fit model)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def Gaussian(x, amp, mu, sigma):
        return  amp/(sigma*np.sqrt(2*np.pi)) * np.exp(-(x-mu)**2/(2*sigma**2))
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak
    model = EMGModel(Gaussian, prefix=pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=0)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg01(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(0,1) emgfit model (single-peak fit model with one exponential 
    tail on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg01(x, amp, mu, sigma, tau_p1):
        return amp*h_emg(x, mu, sigma, 0, (0,),(0,),(1,),(tau_p1,))
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak
    model = EMGModel(emg01, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg10(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(1,0) emgfit model (single-peak fit model with one exponential 
    tail on the left)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg10(x, amp, mu, sigma, tau_m1):
        return amp*h_emg(x, mu, sigma, 1, (1,),(tau_m1,),(0,),(0,))
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak
    model = EMGModel(emg10, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg11(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(1,1) emgfit model (single-peak fit model with one exponential 
    tail on the left and one exponential tail on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu0' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg11(x, amp, mu, sigma, theta, tau_m1, tau_p1):
        return amp*h_emg(x, mu, sigma, theta, (1,),(tau_m1,),(1,),(tau_p1,)) # from emg_funcs.py
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak
    model = EMGModel(emg11, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'theta', value= init_pars['theta'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg12(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(1,2) emgfit model (single-peak fit model with one exponential 
    tail on the left and two exponential tails on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu0' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg12(x, amp, mu, sigma, theta, tau_m1,eta_p1,eta_p2,tau_p1,tau_p2):
        return amp*h_emg(x, mu, sigma, theta, (1,),(tau_m1,),(eta_p1,eta_p2),(tau_p1,tau_p2)) # from emg_funcs.py
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak (e.g. 'p0' for peak with index 0)
    model = EMGModel(emg12, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'theta', value= init_pars['theta'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p1', value= init_pars['eta_p1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p2', value= init_pars['eta_p2'], min=0, max=1, expr='1-'+pref+'eta_p1') # ensures normalization of eta_p's
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p2', value= init_pars['tau_p2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg21(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(2,1) emgfit model (single-peak fit model with two exponential 
    tails on the left and one exponential tail on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg21(x, amp, mu, sigma, theta, eta_m1,eta_m2,tau_m1,tau_m2,tau_p1):
        return amp*h_emg(x, mu, sigma, theta, (eta_m1,eta_m2),(tau_m1,tau_m2),(1,),(tau_p1,)) # from emg_funcs.py
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak (e.g. 'p0' for peak with index 0)
    model = EMGModel(emg21, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'theta', value= init_pars['theta'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m1', value= init_pars['eta_m1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m2', value= init_pars['eta_m2'], min=0, max=1, expr='1-'+pref+'eta_m1') # ensures normalization of eta_m's
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m2', value= init_pars['tau_m2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg22(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(2,2) emgfit model (single-peak fit model with two exponential 
    tails on the left and two exponential tails on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters of the reference peak ('amp' and 'mu'
        parameters in `init_pars` dictionary are updated with the given values
        for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg22(x, amp, mu, sigma, theta, eta_m1,eta_m2,tau_m1,tau_m2,eta_p1,eta_p2,tau_p1,tau_p2):
        return amp*h_emg(x, mu, sigma, theta, (eta_m1,eta_m2),(tau_m1,tau_m2),(eta_p1,eta_p2),(tau_p1,tau_p2)) # from emg_funcs.py
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak (e.g. 'p0' for peak with index 0)
    model = EMGModel(emg22, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)
    
    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'theta', value= init_pars['theta'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m1', value= init_pars['eta_m1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m2', value= init_pars['eta_m2'], min=0, max=1, expr='1-'+pref+'eta_m1') # ensures normalization of eta_m's
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m2', value= init_pars['tau_m2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p1', value= init_pars['eta_p1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p2', value= init_pars['eta_p2'], min=0, max=1, expr='1-'+pref+'eta_p1') # ensures normalization of eta_p's
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p2', value= init_pars['tau_p2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg23(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(2,3) emgfit model (single-peak fit model with two exponential 
    tails on the left and three exponential tails on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg23(x, amp, mu, sigma, theta, eta_m1,eta_m2,tau_m1,tau_m2,eta_p1,eta_p2,eta_p3,delta_p,tau_p1,tau_p2,tau_p3):
        return amp*h_emg(x, mu, sigma, theta, (eta_m1,eta_m2),(tau_m1,tau_m2),(eta_p1,eta_p2,eta_p3),(tau_p1,tau_p2,tau_p3)) # from emg_funcs.py
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak (e.g. 'p0' for peak with index 0)
    model = EMGModel(emg23, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    if init_pars['eta_p2'] < 0: # avoids math errors in peak shape uncertainty estimation
        init_pars['eta_p2'] = 0
    if init_pars['eta_p2'] > 1:
        init_pars['eta_p2'] = 1

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'theta', value= init_pars['theta'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m1', value= init_pars['eta_m1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m2', value= init_pars['eta_m2'], min=0, max=1, expr='1-'+pref+'eta_m1') # ensures normalization of eta_m's
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m2', value= init_pars['tau_m2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p1', value= init_pars['eta_p1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'delta_p', value= init_pars['eta_p1']+init_pars['eta_p2'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p2', value= init_pars['eta_p2'], min=0, max=1, vary=vary_shape_pars, expr= pref+'delta_p-'+pref+'eta_p1')
    model.set_param_hint(pref+'eta_p3', value= 1-init_pars['eta_p1']-init_pars['eta_p2'], min=0, max=1, expr='1-'+pref+'eta_p1-'+pref+'eta_p2') # ensures normalization of eta_p's
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p2', value= init_pars['tau_p2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p3', value= init_pars['tau_p3'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg32(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(3,2) emgfit model (single-peak fit model with three exponential
    tails on the left and two exponential tails on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg32(x, amp, mu, sigma, theta, eta_m1,eta_m2,eta_m3,delta_m,tau_m1,tau_m2,tau_m3,eta_p1,eta_p2,tau_p1,tau_p2):
        return amp*h_emg(x, mu, sigma, theta, (eta_m1,eta_m2,eta_m3),(tau_m1,tau_m2,tau_m3),(eta_p1,eta_p2),(tau_p1,tau_p2)) # from emg_funcs.py
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak (e.g. 'p0' for peak with index 0)
    model = EMGModel(emg32, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    if init_pars['eta_m2'] < 0: # avoids math errors in peak shape uncertainty estimation
        init_pars['eta_m2'] = 0
    if init_pars['eta_m2'] > 1:
        init_pars['eta_m2'] = 1

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'theta', value= init_pars['theta'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m1', value= init_pars['eta_m1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'delta_m', value= init_pars['eta_m1']+init_pars['eta_m2'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m2', value= init_pars['eta_m2'], min=0, max=1, vary=vary_shape_pars, expr= pref+'delta_m-'+pref+'eta_m1')
    model.set_param_hint(pref+'eta_m3', value= 1-init_pars['eta_m1']-init_pars['eta_m2'], min=0, max=1, expr='1-'+pref+'eta_m1-'+pref+'eta_m2')
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m2', value= init_pars['tau_m2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m3', value= init_pars['tau_m3'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p1', value= init_pars['eta_p1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p2', value= 1-init_pars['eta_p1'], min=0, max=1, vary=vary_shape_pars, expr= '1-'+pref+'eta_p1') # ensures normalization of eta_p's
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p2', value= init_pars['tau_p2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


def emg33(peak_index, mu0, amp0, init_pars=pars_dict, vary_shape_pars=True, **kws):
    """
    Hyper-EMG(3,3) emgfit model (single-peak fit model with three exponential
    tails on the left and three exponential tails on the right)

    Parameters
    ----------
    peak_index :  int
        Index of peak to fit.
    mu0 : float
        Initial guess for centroid of underlying Gaussian.
    amp0 : float
        Initial guess for peak amplitude.
    init_pars : dict
        Initial shape parameters ('amp' and 'mu' parameters in `init_pars`
        dictionary are updated with the given values for `amp0` and `mu0`).
    vary_shape_pars : bool
        Whether to vary or fix peak shape parameters (i.e. sigma, theta,
        eta's and tau's).
    kws : Keyword arguments to pass to EMGModel interface.

    Returns
    -------
    :class:`emgfit.model.EMGModel`
        `emgfit` model object

    """
    # Define model function
    def emg33(x, amp, mu, sigma, theta, eta_m1,eta_m2,eta_m3,delta_m,tau_m1,tau_m2,tau_m3,eta_p1,eta_p2,eta_p3,delta_p,tau_p1,tau_p2,tau_p3):
        return amp*h_emg(x, mu, sigma, theta, (eta_m1,eta_m2,eta_m3),(tau_m1,tau_m2,tau_m3),(eta_p1,eta_p2,eta_p3),(tau_p1,tau_p2,tau_p3)) # from emg_funcs.py
    pref = 'p{0}_'.format(peak_index) # set prefix for respective peak (e.g. 'p0' for peak with index 0)
    model = EMGModel(emg33, prefix = pref, nan_policy='propagate', 
                     vary_shape=vary_shape_pars, **kws)

    if init_pars['eta_m2'] < 0: # avoids math errors in peak shape uncertainty estimation
        init_pars['eta_m2'] = 0
    if init_pars['eta_m2'] > 1:
        init_pars['eta_m2'] = 1
    if init_pars['eta_p2'] < 0:
        init_pars['eta_p2'] = 0
    if init_pars['eta_p2'] > 1:
        init_pars['eta_p2'] = 1

    # Add parameters bounds or restrictions and define starting values
    model.set_param_hint(pref+'amp', value=amp0, min=AMP_MIN)
    model.set_param_hint(pref+'mu', value=mu0, min=mu0 - MU_VAR_NSIGMA*init_pars['sigma'], max=mu0 + MU_VAR_NSIGMA*init_pars['sigma'])
    model.set_param_hint(pref+'sigma', value= init_pars['sigma'], min=SIGMA_MIN, max=SIGMA_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'theta', value= init_pars['theta'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m1', value= init_pars['eta_m1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'delta_m', value= init_pars['eta_m1']+init_pars['eta_m2'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_m2', value= init_pars['eta_m2'], min=0, max=1, vary=vary_shape_pars, expr= pref+'delta_m-'+pref+'eta_m1')
    model.set_param_hint(pref+'eta_m3', value= 1-init_pars['eta_m1']-init_pars['eta_m2'], min=0, max=1, expr='1-'+pref+'eta_m1-'+pref+'eta_m2') # ensures normalization of eta_m's
    model.set_param_hint(pref+'tau_m1', value= init_pars['tau_m1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m2', value= init_pars['tau_m2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_m3', value= init_pars['tau_m3'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p1', value= init_pars['eta_p1'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'delta_p', value= init_pars['eta_p1']+init_pars['eta_p2'], min=0, max=1, vary=vary_shape_pars)
    model.set_param_hint(pref+'eta_p2', value= init_pars['eta_p2'], min=0, max=1, vary=vary_shape_pars, expr= pref+'delta_p-'+pref+'eta_p1')
    model.set_param_hint(pref+'eta_p3', value= 1-init_pars['eta_p1']-init_pars['eta_p2'], min=0, max=1, expr='1-'+pref+'eta_p1-'+pref+'eta_p2') # ensures normalization of eta_p's
    model.set_param_hint(pref+'tau_p1', value= init_pars['tau_p1'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p2', value= init_pars['tau_p2'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)
    model.set_param_hint(pref+'tau_p3', value= init_pars['tau_p3'], min=TAU_MIN, max=TAU_MAX_NSIGMA*init_pars['sigma'], vary=vary_shape_pars)

    return model


###############################################################################
###### Define emgfit ConstantModel class
class ConstantModel(EMGModel):
    """Constant background model""" 
    def __init__(self, cost_func='default', independent_vars=['x'], 
                 prefix='bkg', nan_policy='propagate',**kwargs):
        kwargs.update({'prefix': prefix, 'nan_policy': nan_policy,
                       'independent_vars': independent_vars})
        """Create constant background model 

        Parameters 
        ----------
        cost_func : str, optional
            Name of cost function to use for minimization - overrides the 
            model's :attr:`~lmfit.model.Model._residual` attribute. 

            - If ``'chi-square'``, the fit is performed by minimizing Pearson's
              chi-squared statistic:

              .. math::

                  \\chi^2_P = \\sum_i \\frac{(f(x_i) - y_i)^2}{f(x_i)}.

            - If ``'MLE'``, a binned maximum likelihood estimation is performed
              by minimizing the (doubled) negative log likelihood ratio:

              .. math::

                  L = 2\\sum_i \\left[ f(x_i) - y_i + y_i ln\\left(\\frac{y_i}{f(x_i)}\\right)\\right]

            - If ``'default'`` (default), use the standard residual from lmfit.

            See `Notes` of :meth:`~emgfit.spectrum.spectrum.peakfit` method for 
            more details.
        independent_vars : :obj:`list` of :obj:`str`, optional
            Arguments to `func` that are independent variables (default is
            None).
            parameters (default is None).
        prefix : str, optional
            Prefix used for the model.
        nan_policy : {'raise', 'propagate', 'omit'}, optional
            How to handle NaN and missing values in data. See Notes below.
        **kws : dict, optional
            Additional keyword arguments to pass to model function.

        """

        def constant(x, c=0.0):
            return c * np.ones(np.shape(x))
        super().__init__(constant, cost_func=cost_func, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Estimate initial model parameter values from data."""
        pars = self.make_params()

        pars[f'{self.prefix}c'].set(value=data.mean())
        return update_param_vals(pars, self.prefix, **kwargs)



import pytest
import numpy as np
import emgfit as emg
import emgfit.hypothesis_tests as tests
from emgfit.sample import simulate_events

class TestHypothesisTests:
    def test_run_GV_likelihood_ratio_test(self):
        """Test GV likelihood ratio test with simulated data"""
        # Define parameters of reference spectrum
        shape_pars = emg.fit_models.create_default_init_pars(mass_number=100)
        mus = [100.00134, 100.00134 + 900e-06]
        amps = [1., 0.006]
        bkg_c = 1.
        x_range = 0.015
        x_min, x_max = mus[0] - x_range/2, mus[0] + x_range/2
        N_events = 10000
        bin_width = 20e-06
        N_bins = int(np.round((x_max - x_min)/bin_width))

        # Simulate spectrum
        np.random.seed(1543) 
        data = simulate_events(shape_pars, mus, amps, bkg_c, N_events, 
                               x_min, x_max, N_bins=N_bins)
        spec = emg.spectrum(df=data, show_plot=False)
        spec.add_peak(mus[0]+20e-06)

        # Set peak shape to the true peak shape used for the event sampling
        spec.index_shape_calib = 0
        spec.shape_cal_pars = shape_pars
        spec.fit_model = "emg33"

        # Fit data with null-model
        spec.set_lit_values(0, 100.00134, 0.0)
        spec.show_peak_properties()
        spec.fit_peaks(sigmas_of_conf_band=3, index_mass_calib=0, 
                       show_plots=False)

        # Define constants for LRTs
        alt_mu = mus[0] + 910e-06
        alt_x_min = alt_mu -  992e-06
        alt_x_max = alt_mu + 1008e-06 
        GV_seed = 42
        steps = 100
        N_GV_spectra = 5 #100

        # Run GV LRT
        LRT_results = tests.run_GV_likelihood_ratio_test(
                          spec, 0, alt_x_min, alt_x_max, alt_x_steps=steps,
                          min_significance=3, N_spectra=N_GV_spectra, c0=0.5, 
                          seed=GV_seed, show_fits=False, show_upcrossings=False)


        assert np.isclose(LRT_results["LLR"], 15.71, rtol=1e-03, atol=1e-02)
        assert np.isclose(LRT_results["p-value"], 1.23e-03, rtol=1e-03, 
                          atol=1e-05)
        assert np.isclose(LRT_results["p-value error"], 2.00e-04, rtol=1e-03, 
                          atol=1e-05)
        assert LRT_results["reject_null_model"] is True


    def test_MC_likelihood_ratio_test(self):
        """Test MC likelihood ratio test with simulated data"""
        # Define parameters of reference spectrum
        shape_pars = emg.fit_models.create_default_init_pars(mass_number=100)
        mus = [100.00134, 100.00134 + 900e-06]
        amps = [1., 0.006]
        bkg_c = 1.
        x_range = 0.015
        x_min, x_max = mus[0] - x_range/2, mus[0] + x_range/2
        N_events = 10000
        bin_width = 20e-06
        N_bins = int(np.round((x_max - x_min)/bin_width))

        # Simulate spectrum
        np.random.seed(1543) 
        data = simulate_events(shape_pars, mus, amps, bkg_c, N_events, 
                               x_min, x_max, N_bins=N_bins)
        spec = emg.spectrum(df=data, show_plot=False)
        spec.add_peak(mus[0]+20e-06)

        # Set peak shape to the true peak shape used for the event sampling
        spec.index_shape_calib = 0
        spec.shape_cal_pars = shape_pars
        spec.fit_model = "emg33"

        # Fit data with null-model
        spec.set_lit_values(0, 100.00134, 0.0)
        spec.show_peak_properties()
        spec.fit_peaks(sigmas_of_conf_band=3, index_mass_calib=0, 
                       show_plots=False)

        # Define constants for LRTs
        alt_mu = mus[0] + 910e-06
        alt_x_min = alt_mu -  992e-06
        alt_x_max = alt_mu + 1008e-06
        seed = 42
        N_MC_spectra = 5000 
        from emgfit.fit_models import get_mu0
        alt_mu_min = get_mu0(alt_x_min, spec.shape_cal_pars, spec.fit_model)
        alt_mu_max = get_mu0(alt_x_max, spec.shape_cal_pars, spec.fit_model)

        # Run MC LRT
        MC_LRT_results = tests.run_MC_likelihood_ratio_test(
                             spec, 0, alt_mu, alt_mu_min=alt_mu_min, 
                             alt_mu_max=alt_mu_max, N_spectra=N_MC_spectra, 
                             seed=seed, min_significance=3, show_plots=False, 
                             show_results=False, show_LLR_hist=False)

        assert np.isclose(MC_LRT_results["LLR"], 15.71, rtol=1e-03, atol=1e-02)
        assert np.isclose(MC_LRT_results["p-value"], 0.000600120024004,
                          rtol=1e-02, atol=1e-05)
        assert np.isclose(MC_LRT_results["p-value error"], 0.00034647945740525,
                          rtol=1e-02, atol=1e-05)
        assert MC_LRT_results["reject_null_model"] is True 



###################################################################################################
##### Module for importing and handling of data from the Atomic Mass Evaluation
##### (AME) for emgfit package
##### Author: Stefan Paul

from .config import *
from pathlib import Path
import pandas as pd
import numpy as np
from copy import deepcopy

##### Import AME2016 & AME2020 data into pandas dataframes
DIR = Path(__file__).parent  # get directory containing this file
FNAME_AME2016 = str(DIR)+"/AME2016/AME2016-mass-formatted.csv"
DF_AME2016 = pd.read_csv(FNAME_AME2016, encoding='UTF-8', delimiter=';')
DF_AME2016.set_index(['A','Element'], inplace=True)
FNAME_AME2020 = str(DIR)+"/AME2020/AME2020-mass-formatted.csv"
DF_AME2020 = pd.read_csv(FNAME_AME2020, encoding='UTF-8', delimiter=';')
DF_AME2020.set_index(['A','Element'], inplace=True)

##### Define functions
def mdata_AME(El, A, src='AME2020'):
    """Grabs atomic mass data from the atomic mass evaluation (AME).

    Parameters
    ----------
    El : str
        String with element symbol.
    A : int
        Mass number of isotope of interest.
    src : str, optional, default: AME2020
        Source of literature mass data (either 'AME2016' or 'AME2020').

    Returns
    -------
    list (str,int,float,float,bool)
        [Element, Z, A, atomic AME mass [u], atomic AME mass error [u], boolean
        flag for extrapolated AME mass]

    """
    if src == 'AME2020':
        df_AME = deepcopy(DF_AME2020)
    elif src == 'AME2016':
        df_AME = deepcopy(DF_AME2016)

    try:
        Z = df_AME['Z'].loc[(A,El)]
        m_AME = df_AME['ATOMIC MASS [µu]'].loc[(A,El)]*1e-06
        m_AME_error = df_AME['Error ATOMIC MASS [µu]'].loc[(A,El)]*1e-06
        extrapolated_yn = df_AME['Extrapolated?'].loc[(A,El)]
    except KeyError:
        msg = "No AME values found for {0}-{1}.".format(El,A)
        raise Exception(msg) from None

    return [El, Z, A, m_AME, m_AME_error, extrapolated_yn]


def get_El_from_Z(Z):
    """Convenience function to grab element symbol from given proton number.

    Parameters
    ----------
    Z : int or array-like of int
        Proton number.

    Returns
    -------
    str
        Symbol of element with specified proton number.
    """
    # Use only AME2020 since AME2016 contains some misassignments for superheavy
    # element symbols
    if isinstance(Z, (list, tuple, np.ndarray)):
        li = [DF_AME2020.index[(DF_AME2020['Z']==Z_i)].get_level_values(
                                                     'Element')[0] for Z_i in Z]
        El = np.array(li)
    else:
        mask = (DF_AME2020['Z'] == Z)
        El = DF_AME2020.index[mask].get_level_values('Element')[0]
    return El


def splitspecies(s):
    """Splits ion species string into list of constituent atom strings.

    Parameters
    ----------
    s : str
        Input species string.

    Returns
    -------
    list of str
        List of constituent atom strings contained in `s`.

    Examples
    -------
    >>> splitspecies('4H1:1C12')  # returns ``['4H1','1C12']``
    >>> splitspecies('H1:O16')    # returns ``['H1','O16']``

    """
    return s.split(':')


def splitparticle(s):
    """Extracts number, particle/element type and mass number of particle string
    (e.g. 1Cs133, Cs133, 1e).

    Parameters
    ----------
    s : str
        Input particle string.

    Returns
    -------
    tuple of (int, str, int)
        Tuple of signature (number of particles, element symbol, atomic mass number)

    Examples
    --------
    >>> splitspecies('1Cs133') # returns (1,'Cs',133)
    >>> splitspecies('-e')     # returns (-1,'e',0)


    """
    tail = s.lstrip('+-0123456789')
    head = s[:-len(tail)]
    if head == '+' or head == '': # handle omitted 1 or plus sign
        n = int(1)
    elif head == '-': # handle omitted 1
        n = int(-1)
    else:
        n = int(head) # leading number including sign (if present)
    El = tail.rstrip('0123456789') # get central letters
    if El == 'e' and len(El) == len(tail): # handle electron strings, e.g. ':-1e'
        A = 0
    else: # handle hadrons
        A = int(tail[len(El):]) # trailing number
    return n, El, A


def get_charge_state(species):
    """ Return charge state of given species

    Parameters
    ----------
    species : str
        String with name of species.

    Notes
    -----
    `species` strings follow the :ref:`:-notation`.

    The ionic charge state is defined by subtracting the desired number of
    electrons from the atomic species (i.e. ``':-1e'`` for singly charged
    cations, ``':-2e'`` for doubly charged cations etc.).


    """
    if species == '?':
        return None
    z = 0
    for ptype in splitspecies(species): # loop over particle/atom types
        # Remove trailing '?' (flag for tentative species IDs)
        if ptype == '?':
            continue
        else:
            ptype = ptype.rstrip('m? ')
            n, El, A = splitparticle(ptype)
            if El == 'e': # electron
                z = -n  # set charge state

    return z


def get_AME_values(species, Ex=0.0, Ex_error=0.0, src='AME2020'):
    """Calculates the AME mass, AME mass error, the extrapolation flag and the
    mass number A of the given atomic or molecular species.

    Parameters
    ----------
    species : str
        String with name of species to grab AME values for.
    Ex : float [keV], optional, default: 0.0
        Isomer excitation energy in keV to add to ground-state literature
        mass.
    Ex_error : float [keV], optional, default: 0.0
        Uncertainty of isomer excitation energy in keV to add in quadrature
        to ground-state literature mass uncertainty.
    src : str, optional, default: AME2020
        Source of literature data ('AME2016' or 'AME2020').

    Notes
    -----
    `species` strings follow the :ref:`:-notation`.

    Atoms with the nucleus in an isomeric state are flagged by appending an 'm'
    or any of 'm0', 'm1', ..., 'm9' to the atom's substring. For species with
    isomers the literature mass values are only returned if the excitation
    energy is user-specified with the `Ex` argument. In this case, `Ex` is
    added to the AME value for the ground state mass and `Ex_error` is added in
    quadrature to the respective AME uncertainty.

    Both the atomic binding energy of stripped-off electrons as well as the
    uncertainty of the electron mass are neglected in the calculation of the
    ionic AME mass.

    Returns
    -------
    tuple of (float,float,bool,int)
        List containing relevant AME data for `species`:
        (AME mass [u], AME mass uncertainty [u],
        boolean flag for extrapolated species, atomic mass number)

    """
    m = 0.0
    m_error_sq = 0.0
    A_tot = 0
    extrapol = False # initialize boolean flag as False
    isomer_flags = ('m','m0','m1','m2','m3','m4','m5','m6','m7','m8','m9')
    isomer_count = 0 # number of isomeric particle types found
    for ptype in splitspecies(species): # loop over particle/atom types

        # Abort when encountering unidentified species
        if ptype == '?':
            m, m_error, extrapol, A_tot = None, None, False, None
            break # still check warnings below

        # Remove trailing '?' (flag for tentative species IDs)
        ptype = ptype.rstrip('? ')

        # Check for isomer (flagged by 'm' or 'm0' to 'm9' at the end)
        if ptype.endswith(isomer_flags):
            if Ex == 0.0: # abort if no `Ex` is given
                from warnings import warn
                msg = str("{} is labelled as isomer but no excitation energy "
                          "was specified with the `Ex` argument."
                          ).format(species)
                warn(msg)
                return None, None, False, None
            if Ex_error == 0.0:
                from warnings import warn
                msg = str("Uncertainty of isomer excitation energy `Ex_error` "
                          "unspecified for {}.").format(ptype)
                warn(msg)
            isomer_yn = True
            isomer_count += 1
            ptype = ptype.rstrip('0123456789')
            ptype = ptype.rstrip('m')
        else:
            isomer_yn = False

        # Update lit. values
        n, El, A = splitparticle(ptype)
        if El == 'e' and m is not None: # electron
            m += n*m_e
            # neglect uncertainty of m_e
        elif isomer_yn: # isomeric species
            mdata = mdata_AME(El, A, src=src)
            m += n*(mdata[3] + Ex/u_to_keV)
            m_error_sq += (n*mdata[4])**2 + (n*Ex_error/u_to_keV)**2
            m_error = np.sqrt(m_error_sq)
            A_tot += n*A
            if mdata[5]:
                extrapol = True # flag for any extrapolated masses in species
        else: # regular atom
            mdata = mdata_AME(El, A, src=src)
            m += n*(mdata[3])
            m_error_sq += (n*mdata[4])**2
            m_error = np.sqrt(m_error_sq)
            A_tot += n*A
            if mdata[5]:
                extrapol = True # flag for any extrapolated masses in species

    # Issue warnings for isomers if needed
    if Ex != 0.0 and isomer_count == 0: # no isomer flag but Ex given
        from warnings import warn
        msg = str("The specified isomer excitation energy `Ex` and its "
                  "error `Ex_error` are ignored since the species label "
                  "'{}' does not contain any isomer markers.").format(species)
        warn(msg)
    elif isomer_count > 1:
        from warnings import warn
        msg = str("{} contains multiple isomeric constituents. This method "
                  "only supports ions with a single isotopic species in an "
                  "isomeric state.").format(species)
        warn(msg)

    return m, m_error, extrapol, A_tot



################################################################################
##### Python module for creating simulated time-of-flight mass spectra with
##### Gaussian and hyper-exponentially-modified Gaussian lines shapes
##### Author: Stefan Paul

import numpy as np
import pandas as pd
import warnings
from scipy.stats import exponnorm, uniform, norm

################################################################################
##### Define functions for drawing random variates from Gaussian and hyper-EMG
##### PDFs
norm_precision = 1e-06 # required precision for normalization of eta parameters


def Gaussian_rvs(mu, sigma , N_samples=None):
    """Draw random samples from a Gaussian probability density function

    Parameters
    ----------
    mu : float
        Nominal position of simulated peak (mean of Gaussian).
    sigma : float
        Nominal standard deviation of the simulated Gaussian peak.
    N_samples : int, optional, default: 1
        Number of random events to sample.

    Returns
    -------
    :class:`numpy.ndarray` of floats
        Array with simulated events.

    """
    rvs = norm.rvs(loc=mu, scale=sigma, size=N_samples)
    return rvs


def _h_m_i_rvs(mu, sigma, tau_m, N_i):
    """Helper function for definition of h_m_emg_rvs """
    rvs = mu - exponnorm.rvs(loc=0,scale=sigma,K=tau_m/sigma,size=N_i)
    return rvs


def h_m_emg_rvs(mu, sigma, *t_args,N_samples=None):
    """Draw random samples from negative skewed hyper-EMG probability density

    Parameters
    ----------
    mu : float
        Nominal position of simulated peak (mean of underlying Gaussian).
    sigma : float
        Nominal standard deviation (of the underlying Gaussian) of the simulated
        hyper-EMG peak.
    theta : float
        Mixing weight of pos. & neg. skewed EMG distributions.
    t_args : list of lists of float
        List containing lists of the EMG tail parameters with the signature:
        [[eta_m1, eta_m2, ...], [tau_m1, tau_m2, ...]]
    N_samples : int, optional, default: 1
        Number of random events to sample.

    Returns
    -------
    :class:`numpy.ndarray` of floats
        Array with simulated events.

    """
    if not isinstance(N_samples,int):
        raise TypeError("N_samples must be of type int")
    li_eta_m = t_args[0]
    li_tau_m = t_args[1]
    t_order_m = len(li_eta_m) # order of negative tail exponentials
    if abs(sum(li_eta_m) - 1) > norm_precision:
        raise Exception("eta_m's don't add up to 1.")
    if len(li_tau_m) != t_order_m:  # check if all arguments match tail order
        raise Exception("orders of eta_m and tau_m do not match!")

    # randomly distribute ions between tails according to eta_m weights
    tail_nos = np.random.choice(range(t_order_m),size=N_samples,p = li_eta_m)
    rvs = np.array([])
    for i in range(t_order_m):
        N_i = np.count_nonzero(tail_nos == i)
        tau_m = li_tau_m[i]
        rvs_i = _h_m_i_rvs(mu,sigma,tau_m,N_i)
        rvs = np.append(rvs,rvs_i)
    return rvs


def _h_p_i_rvs(mu, sigma, tau_p, N_i):
    """Helper function for definition of h_p_emg_rvs """
    rvs = exponnorm.rvs(loc=mu,scale=sigma,K=tau_p/sigma,size=N_i)
    return rvs


def h_p_emg_rvs(mu, sigma, *t_args, N_samples=None):
    """Draw random samples from pos. skewed hyper-EMG probability density

    Parameters
    ----------
    mu : float
        Nominal position of simulated peak (mean of underlying Gaussian).
    sigma : float
        Nominal standard deviation (of the underlying Gaussian) of the simulated
        hyper-EMG peak.
    t_args : list of lists of float
        List containing lists of the EMG tail parameters with the signature:
        [[eta_p1, eta_p2, ...], [tau_p1, tau_p2, ...]]
    N_samples : int, optional, default: 1
        Number of random events to sample.

    Returns
    -------
    :class:`numpy.ndarray` of floats
        Array with simulated events.

    """
    if not isinstance(N_samples,int):
        raise TypeError("N_samples must be of type int")
    li_eta_p = t_args[0]
    li_tau_p = t_args[1]
    t_order_p = len(li_eta_p) # order of negative tail exponentials
    if abs(sum(li_eta_p) - 1) > norm_precision:
        raise Exception("eta_p's don't add up to 1.")
    if len(li_tau_p) != t_order_p:  # check if all arguments match tail order
        raise Exception("orders of eta_p and tau_p do not match!")

    # randomly distribute ions between tails according to eta_p weights
    tail_nos = np.random.choice(range(t_order_p),size=N_samples,p = li_eta_p)
    rvs = np.array([])
    for i in range(t_order_p):
        N_i = np.count_nonzero(tail_nos == i)
        tau_p = li_tau_p[i]
        rvs_i = _h_p_i_rvs(mu,sigma,tau_p,N_i)
        rvs = np.append(rvs,rvs_i)
    return rvs


def h_emg_rvs(mu, sigma , theta, *t_args, N_samples=None):
    """Draw random samples from a hyper-EMG probability density function

    Parameters
    ----------
    mu : float
        Nominal position of simulated peak (mean of underlying Gaussian).
    sigma : float
        Nominal standard deviation (of the underlying Gaussian) of the simulated
        hyper-EMG peak.
    theta : float
        Mixing weight of pos. & neg. skewed EMG distributions.
    t_args : list of lists of float
        List containing lists of the EMG tail parameters with the signature:
        [[eta_m1, eta_m2, ...], [tau_m1, tau_m2, ...], [eta_p1, eta_p2, ...],
        [tau_p1, tau_p2, ...]]
    N_samples : int, optional, default: 1
        Number of random events to sample.

    Returns
    -------
    :class:`numpy.ndarray` of floats
        Array with simulated events.

    """
    if not isinstance(N_samples,int):
        raise TypeError("N_samples must be of type int")
    li_eta_m = t_args[0]
    li_tau_m = t_args[1]
    li_eta_p = t_args[2]
    li_tau_p = t_args[3]
    if theta == 1:
        rvs = h_m_emg_rvs(mu, sigma, li_eta_m, li_tau_m, N_samples=N_samples)
    elif theta == 0:
        rvs = h_p_emg_rvs(mu, sigma, li_eta_p, li_tau_p, N_samples=N_samples)
    else:
        # randomly distribute ions between h_m_emg and h_p_emg according to
        # left-right-weight theta:
        neg = np.random.choice([1,0],size=N_samples,p = [theta,1-theta])
        N_m = int(np.sum(neg)) #int(np.round(theta*N_samples)) # np.rint(theta*N_samples,dtype=int)
        N_p = N_samples - N_m # np.rint((1-theta)*N_samples,dtype=int)
        rvs_m = h_m_emg_rvs(mu, sigma, li_eta_m, li_tau_m, N_samples=N_m)
        rvs_p = h_p_emg_rvs(mu, sigma, li_eta_p, li_tau_p, N_samples=N_p)
        rvs = np.append(rvs_m,rvs_p)
    return rvs


################################################################################
##### Define functions for simulating events or spectra through random sampling
##### from a reference distribution

def simulate_events(shape_pars, mus, amps, bkg_c, N_events, x_min, x_max,
                    out='hist', scl_facs=None, N_bins=None, bin_cens=None):
    """Create simulated detector events drawn from a user-defined probability
    density function (PDF)

    Events can either be output as a list of single events (mass stamps) or as a
    histogram. In histogram output mode, uniform binning is easily realized by
    specifying the `N_bins` argument. More control over the binning can be
    achieved by parsing the desired bin centers to the `bin_cens` argument (e.g.
    for non-uniform binning).

    Parameters
    ----------
    shape_pars : dict
        Peak-shape parameters to use for sampling. The dictionary must follow
        the structure of the :attr:`~spectrum.shape_cal_pars` attribute of the
        :class:`~emgfit.spectrum.spectrum` class.
    mus : float or list of float
        Nominal peak positions of peaks in simulated spectrum.
    amps : float or list of float [(counts in peak)*(bin width in u)]
        Nominal amplitudes of peaks in simulated spectrum.
    bkg_c : float [counts per bin], optional, default: 0.0
        Nominal amplitude of uniform background in simulated spectrum.
    x_min : float
        Beginning of sampling x-range.
    x_max : float
        End of sampling x-range.
    scl_facs : float or list of float, optional
        Scale factors to use for scaling the scale-dependent shape parameters in
        `shape_pars` to a given peak before sampling events. If `None`, no
        shape-parameter scaling is applied.
    N_events : int, optional, default: 1000
        Total number of events to simulate (signal and background events).
    out : str, optional
        Output format of sampled data. Options:

        - ``'hist'`` for binned mass spectrum (default). The centres of the mass
          bins must be specified with the `bin_cens` argument.
        - ``'array'`` for unbinned array of single ion and background events.

    N_bins : int, optional
        Number of uniform bins to use in ``'hist'`` output mode. The **outer**
        edges of the first and last bin are fixed to the start and end of the
        sampling range respectively (i.e. `x_min` and `x_max`). In between, bins
        are distributed with a fixed spacing of (`x_max`-`x_min`)/`N_bins`.
    bin_cens : :class:`numpy.ndarray`
        Centres of bins to use in ``'hist'`` output mode. This argument
        allows the realization of non-uniform binning. Bin edges are centred
        between neighboring bins. Note: Bins outside the sampling range defined
        with `x_min` and `x_max` will be empty.

    Returns
    -------
    :class:`pandas.Dataframe` or :class:`numpy.ndarray`
       If out='hist' a dataframe with a histogram of the format
       [bin centre, counts in bin] is returned. If out='array' an unbinned
       array with the x-values of single ion or background events is returned.

    Notes
    -----
    Random events are created via custom hyper-EMG extensions of Scipy's
    :meth:`scipy.stats.exponnorm.rvs` method.

    Currently, all simulated peaks have identical width and shape (no re-scaling
    of mass-dependent shape parameters to a peak's mass centroid).

    **Mind the different units for peak amplitudes `amps`
    (<counts in peak> * <bin width in x-axis units>) and the background level
    `bkg_c` (counts per bin).** When spectrum data is simulated counts are
    distributed between the different peaks and the background with probability
    weights `amps` / <bin width in u> and `bkg_c` * <number of bins>,
    respectively. As a consequence, simply changing `N_events` (while keeping
    all other arguments constant), will cause `amps` and `bkg_c` to deviate from
    their nominal units.

    """
    mus = np.atleast_1d(mus)
    amps = np.atleast_1d(amps)
    if len(mus) != len(amps):
        raise Exception("Lengths of `mus` and `amps` arrays must match.")
    if type(N_events) != int:
        raise Exception("`N_events` must be of type int.")
    if (mus < x_min).any() or (mus > x_max).any():
        msg = str("At least one peak centroid in `mus` is outside the sampling range.")
        warnings.warn(msg, UserWarning)
    if scl_facs is None:
        scl_facs = np.ones_like(mus)
    else:
        scl_facs = np.atleast_1d(scl_facs)
        if len(mus) != len(scl_facs):
            raise Exception("Lengths of `mus` and `scl_facs` must match.")

    sample_range = x_max - x_min

    # Get bin parameters
    if N_bins is not None and bin_cens is not None:
        msg =  "Either specify the `N_bins` OR the `bin_cens` argument."
        raise Exception(msg)
    elif bin_cens is not None: # user-defined bins
        N_bins = len(bin_cens)
        bin_edges = np.empty((N_bins+1,))
        spacings = bin_cens[1:] - bin_cens[:-1]
        inner_edges = bin_cens[:-1] + spacings/2
        bin_edges[1:-1] = inner_edges # set inner bin edges
        # Get outer edges
        width_start = bin_cens[1] - bin_cens[0]
        bin_edges[0] = bin_cens[0] - width_start/2 # set first bin edge
        width_end = bin_cens[-1] - bin_cens[-2]
        bin_edges[-1] = bin_cens[-1] + width_end/2 # set last bin edge
        bin_width = (bin_edges[-1] - bin_edges[0])/N_bins # AVERAGE bin width
    elif N_bins is not None: # automatic uniform binning
        bin_edges = np.linspace(x_min, x_max, num=N_bins+1, endpoint=True)
        bin_width = sample_range/N_bins
        bin_cens = bin_edges[:-1] + bin_width/2
    elif out == "array":
        bin_width = 1.
        N_bins = 1.
        pass
    else:
        raise Exception("`N_bins` or `bin_cens` argument must be specified!")

    # Prepare shape parameters
    sigma = shape_pars['sigma']
    li_eta_m = []
    li_tau_m = []
    li_eta_p = []
    li_tau_p = []
    for key, val in shape_pars.items():
        if key.startswith('eta_m'):
            li_eta_m.append(val)
        if key.startswith('tau_m'):
            li_tau_m.append(val)
        if key.startswith('eta_p'):
            li_eta_p.append(val)
        if key.startswith('tau_p'):
            li_tau_p.append(val)
    if len(li_tau_m) == 0 and len(li_tau_p) == 0: # Gaussian
        theta = -1 # flag for below
    else:
        try:
            theta = shape_pars['theta']
        except KeyError:
            pass
        if len(li_eta_m) == 0 and len(li_tau_m) == 1: # emg1X
            li_eta_m = [1]
        elif len(li_eta_m) == 0 and len(li_tau_m) == 0: # emg0X
            theta = 0
        if len(li_eta_p) == 0 and len(li_tau_p) == 1: # emgX1
            li_eta_p = [1]
        elif len(li_eta_p) == 0 and len(li_tau_p) == 0: # emgX0
            theta = 1
    if len(li_eta_m) > 0 and abs(sum(li_eta_m) - 1) > norm_precision:
        msg = "Sum of elements in li_eta_m is not normalized to within {}.".format(norm_precision)
        raise Exception(msg)
    if len(li_eta_m) != len(li_tau_m):
        raise Exception("Lengths of li_eta_m and li_tau_m do not match.")
    if len(li_eta_p) > 0 and abs(sum(li_eta_p) - 1) > norm_precision:
        msg = "Sum of elements in li_eta_p is not normalized to within {}.".format(norm_precision)
        raise Exception(msg)
    if len(li_eta_p) != len(li_tau_p):
        raise Exception("Lengths of li_eta_p and li_tau_p do not match.")

    # Distribute counts over different peaks and background (bkgd)
    # randomly distribute ions using amps and c_bkg as prob. weights
    N_peaks = len(amps)
    counts = np.append(amps/bin_width,bkg_c*N_bins) # cts in each peak & bkgd
    weights = counts/np.sum(counts) # normalized probability weights

    peak_dist = np.random.choice(range(N_peaks+1), size=N_events, p=weights)
    N_bkg = np.count_nonzero(peak_dist == N_peaks) # calc. number of bkgd counts

    events = np.array([])
    # Create & add random samples from each individual peak
    for i in range(N_peaks):
        N_i = np.count_nonzero(peak_dist == i) # get no. of ions in peak
        if theta == -1: # Gaussian
            events_i = Gaussian_rvs(mus[i], sigma*scl_facs[i], N_samples=N_i)
        else: # hyper-EMG
            events_i = h_emg_rvs(mus[i], sigma*scl_facs[i], theta, li_eta_m,
                                 np.array(li_tau_m)*scl_facs[i], li_eta_p,
                                 np.array(li_tau_p)*scl_facs[i], N_samples=N_i)
        events = np.append(events, events_i)

    # Create & add background events
    bkg = uniform.rvs(size=N_bkg, loc=x_min, scale=sample_range)
    events = np.append(events, bkg)

    # Discard events outside of specified sampling range
    events = events[np.logical_and(events >= x_min, events <= x_max)]
    N_discarded = N_events - events.size
    if N_discarded > 0:
        msg = str("{:.0f} simulated events fell outside the specified sampling "
                  "range and were discarded. Peak areas and area ratios might "
                  "deviate from expectation.").format(N_discarded)
        warnings.warn(msg, UserWarning)

    # Return unbinned array of events or dataframe with histogram
    if out == 'array':
        np.random.shuffle(events) # randomize event ordering
        return events
    elif out == 'hist':
        y = np.histogram(events, bins=bin_edges)[0]
        df = pd.DataFrame(data=y, index=bin_cens, columns = ['Counts'])
        df.index.rename('m/z [u]', inplace=True)
        return df


def simulate_spectrum(spec, x_cen=None, x_range=None, mus=None, amps=None,
                      scl_facs=None, bkg_c=None, N_events=None,
                      copy_spec=False):
    """Create a simulated spectrum using the attributes of a reference spectrum

    The peak shape of the sampling probability density function (PDF)
    follows the shape calibration of the reference spectrum (`spec`). By
    default, all other parameters of the sampling PDF are identical to the
    best-fit parameters of the reference spectrum. If desired, the positions,
    amplitudes and number of peaks in the sampling PDF as well as the background
    level can be changed with the `mus`, `amps` and `bkg_c` arguments.

    Parameters
    ----------
    spec : :class:`~emgfit.spectrum.spectrum`
        Reference spectrum object whose best-fit parameters will be used to
        sample from.
    mus : float or list of float, optional
        Nominal peak centres of peaks in simulated spectrum. Defaults to the
        mus of the reference spectrum fit.
    amps : float or list of float [(counts in peak)*(bin width in u)], optional
        Nominal amplitudes of peaks in simulated spectrum. Defaults to the
        amplitudes of the reference spectrum fit.
    scl_facs : float or list of float, optional
        Scale factors to use for scaling the scale-dependent shape parameters in
        `shape_pars` to a given peak before sampling events. Defaults to the
        scale factors asociated with the fit results stored in the reference
        spectrum's :attr:`spectrum.fit_results` attribute.
    bkg_c : float [counts per bin], optional
        Nominal amplitude of uniform background in simulated spectrum. Defaults
        to the c_bkg obtained in the fit of the first peak in the reference
        spectrum.
    x_cen : float, optional
        Center of simulated x-range. Defaults to `x_cen` of `spec`.
    x_range : float, optional
        Covered x-range of simulated spectrum. Defaults to `x_range` of
        `spectrum`.
    N_events : int, optional
        Number of ion events to simulate (including background events). Defaults
        to total number of events in `spec`.
    copy_spec : bool, optional, default: False
        If `False` (default), this function returns a fresh
        :class:`~emgfit.spectrum.spectrum` object created from the simulated
        mass data. If `True`, this function returns an exact copy of `spec` with
        only the :attr:`data` attribute replaced by the new simulated mass data.

    Returns
    -------
    :class:`~emgfit.spectrum.spectrum`
       If `copy_spec = False` (default) a fresh spectrum object holding the
       simulated mass data is returned. If `copy_spec = True`, a copy of the
       reference spectrum `spec` is returned with only the :attr:`data`
       attribute replaced by the new simulated mass data.

    Notes
    -----
    Random events are created via custom Hyper-EMG extensions of Scipy's
    :meth:`scipy.stats.exponnorm.rvs` method.

    The returned spectrum follows the binning of the reference spectrum.

    Mind the different units for peak amplitudes `amps`
    (<counts in peak> * <bin width in x-axis units>) and the background level
    `bkg_c` (counts per bin). When spectrum data is simulated counts are
    distributed between the different peaks and the background with probability
    weights `amps` / <bin width in x-axis units> and `bkg_c` * <number of bins>,
    respectively. As a consequence, simply changing `N_events` (while keeping
    all other arguments constant), will cause `amps` and `bkg_c` to deviate from
    their nominal units.

    """
    if spec.fit_results is [] or None:
        raise Exception("No fit results found in reference spectrum `spec`.")
    if x_cen is None and x_range is None:
        x = spec.data.index.values
        bin_width_start = x[1] - x[0]
        bin_width_end = x[-1] - x[-2]
        x_min = x[0] - bin_width_start/2
        x_max = x[-1] + bin_width_end/2
        indeces = range(len(spec.peaks)) # get peak indeces in sampling range
    else:
        x_min = x_cen - x_range
        x_max = x_cen + x_range
        # Get peak indeces in sampling range:
        peaks = spec.peaks
        indeces = [i for i in range(len(peaks)) if x_min <= peaks[i].x_pos <= x_max]
    if mus is None:
        if len(indeces) == 0:
            msg = str("No peaks in sampling range.")
            warnings.warn(msg, UserWarning)
        mus = []
        for i in indeces:
            result = spec.fit_results[i]
            pref = 'p{0}_'.format(i)
            mus.append(result.best_values[pref+'mu'])
    if amps is None:
        amps = []
        for i in indeces:
            result = spec.fit_results[i]
            pref = 'p{0}_'.format(i)
            amps.append(result.best_values[pref+'amp'])
    if scl_facs is None:
        scl_facs = []
        for i in indeces:
            result = spec.fit_results[i]
            pref = 'p{0}_'.format(i)
            try:
                scl_facs.append(result.params[pref+'scl_fac'])
            except KeyError:
                scl_facs.append(1.0)
    if not (len(mus)==len(amps)==len(scl_facs)):
        msg = "`mus`, `amps` and `scl_facs` argments must have the same length!"
        raise Exception(msg)
    if bkg_c is None:
        if len(indeces) == 0:
            raise Exception("Zero background and no peaks in sampling range.")
        result = spec.fit_results[indeces[0]]
        bkg_c = result.best_values['bkg_c']
    if N_events is None:
        # Get total number of counts in simulated region of original spectrum:
        N_events = int(spec.data[x_min:x_max]["Counts"].sum())

    # Create histogram with Monte Carlo events
    x = spec.data[x_min:x_max].index.values
    df = simulate_events(spec.shape_cal_pars, mus, amps, bkg_c, N_events,
                         x_min, x_max, out='hist', scl_facs=scl_facs,
                         N_bins=None, bin_cens=x)

    # Copy original spectrum and overwrite data
    # This copies all results such as peak assignments, PS calibration,
    # fit_results etc.
    if copy_spec:
        from copy import deepcopy
        new_spec = deepcopy(spec)
        new_spec.data = df
    else: # Define a fresh spectrum with sampled data
        from emgfit.spectrum import spectrum
        new_spec = spectrum(df=df, show_plot=False)

    return new_spec


def fit_simulated_spectra(spec, fit_result, alt_result=None, N_spectra=1000,
                          seed=None, randomize_ref_mus_and_amps=False,
                          MC_shape_par_samples=None, n_cores=-1):
    """Fit spectra simulated via sampling from a reference distribution

    This function performs fits of many simulated spectra. The simulated spectra
    are created by sampling events from the best-fit PDF asociated with
    `fit_result` (as e.g. needed for a parametric bootstrap). The `alt_model`
    can be used to perform the fits with a different distribution than the
    reference PDF used for the event sampling.

    Parameters
    ----------
    spec : :class:`emgfit.spectrum.spectrum`
        Spectrum object to perform bootstrap on.
    fit_result : :class:`lmfit.model.ModelResult`
        Fit result object holding the best-fit distribution to sample from.
    alt_result : :class:`lmfit.model.ModelResult`, optional
        Fit result object holding a prepared fit model to be used for the
        fitting. Defaults to the fit model stored in `fit_result`.
    N_spectra : int, optional
        Number of simulated spectra to fit. Defaults to 1000, which
        typically yields statistical uncertainty estimates with a Monte Carlo
        uncertainty of a few percent.
    randomize_ref_mus_and_amps : bool, default: False
        If `True`, the peak and background amplitudes and the peak centroids of
        the reference spectrum to sample from will be varied assuming normal
        distributions around the best-fit values with standard deviations given
        by the respective standard errors stored in `fit_result`.
    MC_shape_par_samples : :class:`pandas.DataFrame` 
        Monte Carlo shape parameter samples to use in the fitting. 
    seed : int, optional
        Random seed to use for reproducible sampling.
    n_cores : int, optional
        Number of CPU cores to use for parallelized fitting of simulated
        spectra. When set to `-1` (default) all available cores are used.

    Returns
    -------
    :class:`numpy.ndarray` of :class:`lmfit.minimizer.MinimizerResult`
        MinimizerResults obtained in the fits of the simulated spectra.

    Note
    ----
    The `randomize_ref_mus_and_amps` option allows one to propagate systematic
    uncertainties in the determination of the reference parameters into the
    Monte Carlo results. If varying the centroid and amplitude parameters of the
    reference spectrum, the standard deviations of the parameter distributions
    will be taken as the respective standard errors determined by lmfit (see
    lmfit fit report table) and might not be consistent with the area and mass
    uncertainty estimates shown in the peak properties table.

    See also
    --------
    :meth:`emgfit.spectrum.spectrum.get_errors_from_resampling`
    :func:`emgfit.sample.simulate_events` for details on the event sampling.

    """
    bkg_c = fit_result.best_values['bkg_c']
    bkg_c_err = fit_result.params['bkg_c'].stderr
    method = fit_result.method
    shape_pars = spec.shape_cal_pars
    x_cen = fit_result.x_fit_cen
    x_range = fit_result.x_fit_range
    x = fit_result.x
    y = fit_result.y
    x_min = x_cen - 0.5*x_range
    x_max = x_cen + 0.5*x_range

    if alt_result is None:
        fit_model = fit_result.fit_model
        model = fit_result.model
        init_pars = fit_result.init_params
    else:
        fit_model = alt_result.fit_model
        model = alt_result.model
        init_pars = alt_result.init_params

    # Determine tail order of fit model for normalization of initial etas
    if fit_result.fit_model.startswith('emg'):
        n_ltails = int(fit_result.fit_model.lstrip('emg')[0])
        n_rtails = int(fit_result.fit_model.lstrip('emg')[1])
    else:
        n_ltails = 0
        n_rtails = 0

    # Collect aLL peaks, peak centroids and amplitudes of fit_result
    fitted_peaks = [idx for idx, p in enumerate(spec.peaks)
                    if x_min < p.x_pos < x_max] # indeces of all fitted peaks
    mus, mu_errs = [], []
    amps, amp_errs = [], []
    scl_facs = []
    for idx in fitted_peaks:
        pref = 'p{0}_'.format(idx)
        mus.append(fit_result.best_values[pref+'mu'])
        mu_errs.append(fit_result.params[pref+'mu'].stderr)
        amps.append(fit_result.best_values[pref+'amp'])
        amp_errs.append(fit_result.params[pref+'amp'].stderr)
        try:
            scl_facs.append(fit_result.params[pref+'scl_fac'])
        except KeyError:
            scl_facs.append(1)

    import emgfit.fit_models as fit_models
    from .model import save_model, load_model
    from lmfit.minimizer import minimize
    import lmfit
    import time
    datetime = time.localtime() # get current date and time
    datetime_str = time.strftime("%Y-%m-%d_%H-%M-%S", datetime)
    if spec.input_filename is not None:
        data_fname = spec.input_filename.rsplit('.', 1)[0] # del. extension
    else:
        data_fname = ''
    modelfname = data_fname+datetime_str+"_resampl_model.sav"
    save_model(model, modelfname)
    N_events = int(np.sum(y))
    funcdefs = {'constant': lmfit.models.ConstantModel,
                str(fit_model): getattr(fit_models,fit_model)}
    def refit(seed, shape_pars_i):
        # create simulated spectrum data by sampling from fit-result PDF
        np.random.seed(seed)
        if randomize_ref_mus_and_amps:
            bkg_c_i = np.random.normal(loc=bkg_c, scale=bkg_c_err, size=1)
            amps_i = np.random.normal(loc=amps, scale=amp_errs)
            mus_i = np.random.normal(loc=mus, scale=mu_errs)
        else:
            bkg_c_i = bkg_c
            amps_i = amps
            mus_i = mus

        if MC_shape_par_samples is not None:
            # Calculate missing parameters from normalization
            if n_ltails == 2:
                shape_pars_i['eta_m2'] = 1 - shape_pars_i['eta_m1']
            elif n_ltails == 3:
                eta_m2 = shape_pars_i['delta_m'] - shape_pars_i['eta_m1']
                shape_pars_i['eta_m3'] = 1 - shape_pars_i['eta_m1'] - eta_m2
            if n_rtails == 2:
                shape_pars_i['eta_p2'] = 1 - shape_pars_i['eta_p1']
            elif n_rtails == 3:
                eta_p2 = shape_pars_i['delta_p'] - shape_pars_i['eta_p1']
                shape_pars_i['eta_p3'] = 1 - shape_pars_i['eta_p1'] - eta_p2
        df = simulate_events(shape_pars_i, mus_i, amps_i, bkg_c_i,
                             N_events, x_min, x_max,
                             out='hist', scl_facs=scl_facs, bin_cens=x)
        new_x = df.index.values
        new_y = df['Counts'].values
        new_y_err = np.maximum(1, np.sqrt(new_y)) # Poisson (counting) stats
        # Weights for residuals: residual = (fit_model - y) * weights
        new_weights = 1./new_y_err

        model = load_model(modelfname, funcdefs=funcdefs)

        # re-perform fit on simulated spectrum - for performance, use only the
        # underlying Minimizer object instead of full lmfit model interface
        try:
            min_res = minimize(model._residual, init_pars, method=method,
                               args=(new_y, new_weights), kws={'x': new_x},
                               scale_covar=False, nan_policy='propagate',
                               reduce_fcn=None, calc_covar=False)
            return min_res

        except ValueError:
            with warnings.catch_warnings():
                warnings.simplefilter('always')
                msg = str("Fit failed with ValueError (likely NaNs in "
                           "y-model array) and will be excluded.")
                warnings.warn(msg, UserWarning)
            return None

    # For reproducible sampling with joblib parallel, generate `N_spectra`
    # random seeds for refit() - only works with the default Loky backend
    np.random.seed(seed=seed) # for reproducible sampling
    joblib_seeds = np.random.randint(2**31, size=N_spectra)
    from tqdm.auto import tqdm # add progress bar with tqdm
    from joblib import Parallel, delayed
    try:
        if MC_shape_par_samples is None:
            min_results = np.array(Parallel(n_jobs=n_cores)
                                    (delayed(refit)(s, shape_pars) for s in tqdm(joblib_seeds)))
        else:
            from .spectrum import _strip_prefs
            min_results = np.array(Parallel(n_jobs=n_cores)
                                    (delayed(refit)(s, _strip_prefs(dict(MC_shape_par_samples.iloc[i]))) for i, s in tqdm(enumerate(joblib_seeds))))
    finally:
        # Force workers to shut down and clean up temp SAV file
        from joblib.externals.loky import get_reusable_executor
        import os
        get_reusable_executor().shutdown(wait=True)
        os.remove(modelfname)

    return min_results

################################################################################
##### Define functions for non-parametric resampling
def resample_events(df, N_events=None, x_cen=None, x_range=0.02, out='hist'):
    """Create simulated spectrum via non-parametric resampling from `df`.

    The simulated data is obtained through resampling from the specified dataset
    with replacement.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        Original histogrammed spectrum data to re-sample from.
    N_events : int, optional
        Number of events to create via non-parametric re-sampling, defaults to
        number of events in original DataFrame `df`.
    x_cen : float [u/z], optional
        Center of mass range to re-sample from. If ``None``, re-sample from
        full mass range of input data `df`.
    x_range : float [u/z], optional
        Width of mass range to re-sample from. Defaults to 0.02 u. `x_range`
        is only relevant if a `x_cen` argument is specified.
    out : str, optional
        Output format of sampled data. Options:

        - ``'hist'`` for binned mass spectrum (default). The centres of the mass
          bins must be specified with the `bin_cens` argument.
        - ``'array'`` for unbinned array of single ion and background events.

    Returns
    -------
    :class:`pandas.Dataframe` or :class:`numpy.ndarray`
       If out='hist' a dataframe with a histogram of the format
       [bin centre, counts in bin] is returned. If out='array' an unbinned
       array with the x-values of single ion or background events is returned.

    """
    if x_cen:
        x_min = x_cen - 0.5*x_range
        x_max = x_cen + 0.5*x_range
        df = df[x_min:x_max]
    mass_bins = df.index.values
    counts = df['Counts'].values.astype(int)

    # Convert histogrammed spectrum (equivalent to MAc HIST export mode) to
    # list of events (equivalent to MAc LIST export mode)
    orig_events =  np.repeat(mass_bins, counts, axis=0)

    # Create new DataFrame of events by bootstrapping from `orig_events`
    if N_events == None:
        N_events = len(orig_events)
    random_indeces = np.random.randint(0, len(orig_events), N_events)
    events = pd.DataFrame(orig_events[random_indeces])

    # Convert list of events back to a DataFrame with histogram data
    bin_cens = df.index.values
    bin_width = df.index.values[1] - df.index.values[0]
    bin_edges = np.append(bin_cens-0.5*bin_width,
                          bin_cens[-1]+0.5*bin_width)

    # Return unbinned array of events or dataframe with histogram
    if out == 'array':
        np.random.shuffle(events) # randomize event ordering
        return events
    elif out == 'hist':
        hist = np.histogram(events, bins=bin_edges)
        df_new = pd.DataFrame(data=hist[0], index=bin_cens, dtype=float,
                            columns=["Counts"])
        df_new.index.name = "m/z [u]"
        return df_new
    


import pytest
import emgfit
import numpy as np

@pytest.fixture
def test_sampling():
    """Helper function to check result of simulate_events() against target PDF
    """
    def test_sampling_func(modname, mu, shape_pars, bkg_c=0, scl_fac=1.0,
                           N_events=10000000):
        x_min = mu - 10*shape_pars["sigma"]
        x_max = mu + 10*shape_pars["sigma"]

        # Simulate events
        from emgfit.sample import simulate_events
        N_bins = int((x_max-x_min)/(0.1*shape_pars["sigma"]))
        bin_width = (x_max - x_min)/N_bins
        amp = N_events*bin_width

        np.random.seed(0)
        data = simulate_events(shape_pars, mu, amp, bkg_c, N_events, x_min,
                               x_max, scl_facs=scl_fac, out='hist',
                               N_bins=N_bins)

        # Check simulated events against generating PDF
        x = data.index.values
        y_sim = data.values.flatten()

        import emgfit.fit_models as fit_models
        modfunc = getattr(fit_models, modname)
        scaled_pars = fit_models.scl_init_pars(shape_pars, scl_coeff=scl_fac)
        mod = modfunc(0, mu, amp, init_pars=scaled_pars, vary_shape_pars=True)
        pars = mod.make_params()
        y_PDF = mod.eval(pars,x=x) # target distribution

        allclose = np.allclose(y_sim, y_PDF, atol=15, rtol=15e-01)
        msg0 = "y_sim & y_PDF differ by >15% and >15counts in some bins."
        assert allclose, msg0

        mean_residual = np.mean(y_PDF - y_sim)
        msg1 = "The mean of y_sim - y_PDF differs from 0 by more than 1."
        assert abs(mean_residual) <= 1, msg1

    return test_sampling_func

class TestSampling():

    def test_Gaussian_sampling(self, test_sampling):
        """Test simulate_events() for sampling from Gaussian PDF"""
        modname = "Gaussian"
        mu = 1000.151
        sigma = 0.5
        shape_pars = {'sigma': sigma}
        test_sampling(modname, mu, shape_pars)

    def test_emg10_sampling(self, test_sampling):
        modname = "emg10"
        mu = 100.1
        sigma = 0.0001
        tau_m1 = 90e-06
        shape_pars = {'sigma': sigma,
                      'tau_m1': tau_m1}
        test_sampling(modname, mu, shape_pars)

    def test_emg21_sampling(self, test_sampling):
        """Test simulate_events() for sampling from emg21 PDF with scaling"""
        modname = "emg21"
        mu = 10.1
        sigma = 0.0005
        theta = 0.41
        eta_m1 = 0.9
        eta_m2 = 1 - eta_m1
        tau_m1 = 900e-06
        tau_m2 = 2000e-06
        tau_p1 = 320e-06
        shape_pars = {'sigma': sigma,
                      'theta': theta,
                      'eta_m1': eta_m1,
                      'eta_m2': eta_m2,
                      'tau_m1': tau_m1,
                      'tau_m2': tau_m2,
                      'tau_p1': tau_p1}
        scl_fac = 0.4
        test_sampling(modname, mu, shape_pars, scl_fac=0.4)


    def test_emg33_sampling(self, test_sampling):
        """Test simulate_events() for sampling from emg33 PDF"""
        modname = "emg33"
        mu = 100.17654
        sigma = 0.005
        theta = 0.81
        eta_m1 = 0.6
        eta_m2 = 0.2
        eta_m3 = 1 - eta_m1 - eta_m2
        tau_m1 = 30e-06
        tau_m2 = 870e-06
        tau_m3 = 2500e-06
        eta_p1 = 0.8
        eta_p2 = 0.15
        eta_p3 = 1 - eta_p1 - eta_p2
        tau_p1 = 320e-06
        tau_p2 = 691e-06
        tau_p3 = 2100e-06
        shape_pars = {'sigma': sigma,
                      'theta': theta,
                      'eta_m1': eta_m1,
                      'eta_m2': eta_m2,
                      'eta_m3': eta_m3,
                      'tau_m1': tau_m1,
                      'tau_m2': tau_m2,
                      'tau_m3': tau_m3,
                      'eta_p1': eta_p1,
                      'eta_p2': eta_p2,
                      'eta_p3': eta_p3,
                      'tau_p1': tau_p1,
                      'tau_p2': tau_p2,
                      'tau_p3': tau_p3}
        test_sampling(modname, mu, shape_pars)

    def test_simulate_spectrum(self):
        # Define spectrum parameters
        modname = "emg21"
        sigma = 0.000045
        theta = 0.41
        eta_m1 = 0.7
        eta_m2 = 1 - eta_m1
        tau_m1 = 30e-06
        tau_m2 = 300e-06
        tau_p1 = 120e-06
        shape_pars = {'sigma':sigma,
                      'theta': theta,
                      'eta_m1': eta_m1,
                      'eta_m2': eta_m2,
                      'tau_m1': tau_m1,
                      'tau_m2': tau_m2,
                      'tau_p1': tau_p1}

        mus = [57, 57.0012]
        x_min = mus[0] - 20*shape_pars["sigma"]
        x_max = mus[-1] + 20*shape_pars["sigma"]
        N_events = 10000000
        bkg_c = 1.5e03

        # Simulate events, create reference spectrum and fit
        from emgfit.sample import simulate_events, simulate_spectrum
        np.random.seed(49)
        N_bins = int((x_max-x_min)/(0.1*shape_pars["sigma"]))
        bin_width = (x_max - x_min)/N_bins
        N_bkg = 0.4*N_bins
        amps = np.array([0.3,0.7])*(N_events-N_bkg)*bin_width
        data = simulate_events(shape_pars, mus, amps, bkg_c, N_events, x_min,
                               x_max, out='hist', N_bins=N_bins)
        spec = emgfit.spectrum(df=data, show_plot=False)
        spec.add_peak(57, verbose= False)
        spec.add_peak(57.0012, verbose = False)
        m, m_err = 57, 1e-06
        spec.set_lit_values(0, m, m_err, verbose = False)
        spec.determine_peak_shape(index_shape_calib=1, x_fit_range=20*sigma,
                                  fit_model=modname, vary_tail_order=False,
                                  show_plots=False)
        spec.fit_peaks(index_mass_calib=0, show_plots=False)

        # Create new simulated spectrum from reference spectrum
        new_spec = simulate_spectrum(spec, copy_spec=False)

        data_ref = spec.data
        data_sim = new_spec.data
        msg0 = "Bin positions x_sim and x_ref disagree by >1e-12."
        assert np.allclose(data_ref.index,data_sim.index, rtol=1e-12), msg0
        msg1 = "Counts y_sim and y_ref deviate by >15% in some bins."
        assert np.allclose(data_ref.values, data_sim.values, rtol=0.15), msg1



##################################################################################
##### Module defining hyper-EMG model interface based on parent classes from lmfit 
##### Authors: lmfit authors and Stefan Paul
import lmfit
import operator
import json
import warnings
import numpy as np
import lmfit.lineshapes as lineshapes
from lmfit.jsonutils import HAS_DILL
from copy import deepcopy
from lmfit.model import _align, tiny
from lmfit import Parameter, Parameters
from lmfit.jsonutils import HAS_DILL, decode4js, encode4js
from numpy import sqrt, log

EPS = 1e-10 # small number to bound Pearson cost function weights

# Use pandas.isnull for aligning missing data if pandas is available.
# otherwise use numpy.isnan
try:
    from pandas import Series, isnull
except ImportError:
    isnull = np.isnan
    Series = type(NotImplemented)

def coerce_arraylike(x):
    """
    coerce lists, tuples, and pandas Series, hdf5 Groups, etc to an
    ndarray float64 or complex128, but leave other data structures
    and objects unchanged

    Function adopted from :mod:`lmfit.model` module.
    
    """
    if isinstance(x, (list, tuple, Series)) or hasattr(x, '__array__'):
        if np.isrealobj(x):
            return np.asfarray(x)
        if np.iscomplexobj(x):
            return np.asfarray(x, dtype=np.complex128)
    return x


class EMGModel(lmfit.model.Model):
    """Create hyper-EMG fit model  
    
    This class inherits from the :class`lmfit.model.Model` class and creates a 
    single-peak model. This class enables overriding of lmfit's default 
    residuals with emgfit's custom cost functions, thereby enabling fits beyond 
    standard least squares minimization. 
        
    """
    def __init__(self, func, cost_func="default", 
                 par_hint_args={}, vary_baseline=True, vary_shape=True, 
                 independent_vars=['x'], param_names=None, 
                 nan_policy='propagate', prefix='', name=None, **kws):
        """
        The model function will normally take an independent variable
        (generally, the first argument) and a series of arguments that are
        meant to be parameters for the model. It will return an array of
        data to model some data as for a curve-fitting problem.

        Method adopted from :mod:`lmfit.model` module.

        Parameters
        ----------
        func : callable
            Function to be wrapped.
        cost_func : str, optional
            Name of cost function to use for minimization - overrides the 
            model's :attr:`~lmfit.model.Model._residual` attribute. 

            - If ``'chi-square'``, the fit is performed by minimizing Pearson's
              chi-squared statistic:

              .. math::

                  \\chi^2_P = \\sum_i \\frac{(f(x_i) - y_i)^2}{f(x_i)}.

            - If ``'MLE'``, a binned maximum likelihood estimation is performed
              by minimizing the (doubled) negative log likelihood ratio:

              .. math::

                  L = 2\\sum_i \\left[ f(x_i) - y_i + y_i ln\\left(\\frac{y_i}{f(x_i)}\\right)\\right]

            - If ``'default'`` (default), use the standard residual from lmfit.

            See `Notes` of :meth:`~emgfit.spectrum.spectrum.peakfit` method for 
            more details.
        par_hint_args : dict of dicts, optional
            Arguments to pass to :meth:`lmfit.model.Model.set_param_hint` to
            modify or add model parameters. The keys of the `par_hint_args`
            dictionary specify parameter names; the values must likewise be
            dictionaries that hold the respective keyword arguments to pass to
            :meth:`~lmfit.model.Model.set_param_hint`.
        vary_baseline : bool, optional, default: `True`
            If `True`, the constant background will be fitted with a varying
            uniform baseline parameter `bkg_c`.
            If `False`, the baseline parameter `bkg_c` will be fixed to 0.
        vary_shape : bool, optional, default: `False`
            If `False` peak-shape parameters of hyper-EMG models (`sigma`, 
            `theta`,`etas` and `taus`) are kept fixed at their initial values. 
            If `True` the shared shape parameters are varied (ensuring 
            identical shape parameters for all peaks).
        independent_vars : :obj:`list` of :obj:`str`, optional
            Arguments to `func` that are independent variables (default is
            None).
        param_names : :obj:`list` of :obj:`str`, optional
            Names of arguments to `func` that are to be made into
            parameters (default is None).
        nan_policy : {'raise', 'propagate', 'omit'}, optional
            How to handle NaN and missing values in data. See Notes below.
        prefix : str, optional
            Prefix used for the model.
        name : str, optional
            Name for the model. When None (default) the name is the same
            as the model function (`func`).
        **kws : dict, optional
            Additional keyword arguments to pass to model function.

        Notes
        -----
        1. Parameter names are inferred from the function arguments, and a
        residual function is automatically constructed.

        2. The model function must return an array that will be the same
        size as the data being modeled.

        3. `nan_policy` sets what to do when a NaN or missing value is
        seen in the data. Should be one of:

           - `'raise'` : raise a `ValueError` (default)
           - `'propagate'` : do nothing
           - `'omit'` : drop missing data

        Examples
        --------
        The model function will normally take an independent variable
        (generally, the first argument) and a series of arguments that are
        meant to be parameters for the model. Thus, a simple peak using a
        Gaussian defined as:

        >>> import numpy as np
        >>> def gaussian(x, amp, cen, wid):
        ...     return amp * np.exp(-(x-cen)**2 / wid)

        can be turned into a Model with:

        >>> gmodel = Model(gaussian)

        this will automatically discover the names of the independent
        variables and parameters:

        >>> print(gmodel.param_names, gmodel.independent_vars)
        ['amp', 'cen', 'wid'], ['x']

        """        
        super().__init__(func, independent_vars=independent_vars, 
                         param_names=param_names, nan_policy=nan_policy, 
                         prefix=prefix, name=name, **kws)
        self._override_residual(cost_func)
        self.par_hint_args = par_hint_args
        self.vary_baseline = vary_baseline
        self.vary_shape = vary_shape


    def _override_residual(self, cost_func):
        """Override residual method 
        
        Parameters
        ----------
        cost_func : str, optional 
            Name of cost function to use for minimization - overrides the 
            model's :attr:`~lmfit.model.Model._residual` attribute. 

            - If ``'chi-square'``, the fit is performed by minimizing Pearson's
              chi-squared statistic:

              .. math::

                  \\chi^2_P = \\sum_i \\frac{(f(x_i) - y_i)^2}{f(x_i)}.

            - If ``'MLE'``, a binned maximum likelihood estimation is performed
              by minimizing the (doubled) negative log likelihood ratio:

              .. math::

                  L = 2\\sum_i \\left[ f(x_i) - y_i + y_i ln\\left(\\frac{y_i}{f(x_i)}\\right)\\right]

            - If ``'default'`` (default), use the standrad residual from lmfit.

            See `Notes` of :meth:`~emgfit.spectrum.spectrum.peakfit` method for 
            more details.

        """
        if cost_func == "default":
            pass
        elif cost_func == "chi-square":
            # Pearson's chi-squared cost function with iterative weights 1/Sqrt(f(x_i))
            def resid_Pearson_chi_square(pars, y_data, weights, **kwargs):
                y_m = self.eval(pars, **kwargs)
                # Calculate weights for current iteration, add tiny number `EPS`
                # in denominator for numerical stability
                weights = 1./sqrt(y_m + EPS)
                return (y_m - y_data)*weights
            # Override lmfit's standard least square residuals with iterative
            # residuals for Pearson chi-square fit
            self._residual = resid_Pearson_chi_square
        elif cost_func == "MLE":
            # Define sqrt of (doubled) negative log-likelihood ratio (NLLR) summands
            def sqrt_NLLR(pars, y_data, weights, **kwargs):
                y_m = self.eval(pars, **kwargs) 
                # Add tiniest pos. float representable by numpy to arguments of
                # np.log to smoothly handle divergences for log(arg -> 0)
                NLLR = 2*(y_m-y_data) + 2*y_data*(log(y_data+tiny)-log(y_m+tiny))
                ret = sqrt(NLLR)
                return ret
            # Override lmfit's standard least square residual with the
            # square-roots of the NLLR summands, this enables usage of scipy's
            # `least_squares` minimizer and yields much faster optimization
            # than with scalar minimizers
            self._residual = sqrt_NLLR    
        else:
            raise Exception(" Definition of `cost_func` failed!")
        self.cost_func = cost_func


    def _reprstring(self, long=False):
        """Print representation string

        Method adopted from :mod:`lmfit.model` module.
        """
        out = self._name
        opts = []
        if len(self._prefix) > 0:
            opts.append(f"prefix='{self._prefix}'")
        if long:
            for k, v in self.opts.items():
                opts.append(f"{k}='{v}'")
        if len(opts) > 0:
            out = f"{out}, {', '.join(opts)}"
        return f"EMGModel({out})"
        

    def _get_state(self):
        """Save a Model for serialization.

        Note: like the standard-ish '__getstate__' method but not really
        useful with Pickle.

        Method modified from :mod:`lmfit.model` module.

        """
        funcdef = None
        if HAS_DILL:
            funcdef = self.func
        if self.func.__name__ == '_eval':
            funcdef = self.expr
        state = (self.func.__name__, funcdef, self.cost_func, 
                 self.par_hint_args, self.vary_baseline, 
                 self.vary_shape, self._name, self._prefix,
                 self.independent_vars, self._param_root_names,
                 self.param_hints, self.nan_policy, self.opts)
        return (state, None, None)
    

    def _set_state(self, state, funcdefs=None):
        """Restore Model from serialization.

        Note: like the standard-ish '__setstate__' method but not really
        useful with Pickle.

        Method adopted from :mod:`lmfit.model` module.

        Parameters
        ----------
        state
            Serialized state from `_get_state`.
        funcdefs : dict, optional
            Dictionary of function definitions to use to construct Model.

        """
        return _buildmodel(state, funcdefs=funcdefs)
    

    def __repr__(self):
        """Return representation of EMGModel."""
        return f"<emgfit.EMGModel: {self.name}>"
    

    def fit(self, data, params=None, weights=None, fitted_peaks=None, 
            method='least_squares', iter_cb=None, scale_covar=True, verbose=False, 
            fit_kws=None, nan_policy=None, calc_covar=True, max_nfev=None,
            coerce_farray=True, **kwargs):
        """Fit the model to the data using the supplied Parameters.

        Method modified from :mod:`lmfit.model` module.

        Parameters
        ----------
        data : array_like
            Array of data to be fit.
        params : Parameters, optional
            Parameters to use in fit (default is None).
        weights : array_like, optional
            Weights to use for the calculation of the fit residual [i.e.,
            `weights*(data-fit)`]. Default is None; must have the same size as
            `data`.
        fitted_peaks : list of :class:`emgfit.spectrum.peak`
            List of peaks to be fitted.
        method : str, optional
            Name of fitting method to use (default is `'least_squares'`).
        iter_cb : callable, optional
            Callback function to call at each iteration (default is None).
        scale_covar : bool, optional
            Whether to automatically scale the covariance matrix when
            calculating uncertainties (default is True).
        verbose : bool, optional
            Whether to print a message when a new parameter is added
            because of a hint (default is True).
        fit_kws : dict, optional
            Options to pass to the minimizer being used.
        nan_policy : {'raise', 'propagate', 'omit'}, optional
            What to do when encountering NaNs when fitting Model.
        calc_covar : bool, optional
            Whether to calculate the covariance matrix (default is True)
            for solvers other than `'leastsq'` and `'least_squares'`.
            Requires the ``numdifftools`` package to be installed.
        max_nfev : int or None, optional
            Maximum number of function evaluations (default is None). The
            default value depends on the fitting method.
        coerce_farray : bool, optional
            Whether to coerce data and independent data to be ndarrays
            with dtype of float64 (or complex128).  If set to False, data
            and independent data are not coerced at all, but the output of
            the model function will be. (default is True)
        **kwargs : optional
            Arguments to pass to the model function, possibly overriding
            parameters.

        Returns
        -------
        :class:`emgfit.model.EMGModelResult`
            EMGModelResult

        Notes
        -----
        1. if `params` is None, the values for all parameters are expected
        to be provided as keyword arguments. Mixing `params` and
        keyword arguments is deprecated (see `Model.eval`).

        2. all non-parameter arguments for the model function, **including
        all the independent variables** will need to be passed in using
        keyword arguments.

        3. Parameters are copied on input, so that the original Parameter objects
        are unchanged, and the updated values are in the returned `EMGModelResult`.

        Examples
        --------
        Take ``t`` to be the independent variable and data to be the curve
        we will fit. Use keyword arguments to set initial guesses:

        >>> result = my_model.fit(data, tau=5, N=3, t=t)

        Or, for more control, pass a Parameters object.

        >>> result = my_model.fit(data, params, t=t)

        """
        if params is None:
            params = self.make_params(verbose=verbose)
        else:
            params = deepcopy(params)

        # If any kwargs match parameter names, override params.
        param_kwargs = set(kwargs.keys()) & set(self.param_names)
        for name in param_kwargs:
            p = kwargs[name]
            if isinstance(p, Parameter):
                p.name = name  # allows N=Parameter(value=5) with implicit name
                params[name] = deepcopy(p)
            else:
                params[name].set(value=p)
            del kwargs[name]

        # All remaining kwargs should correspond to independent variables.
        for name in kwargs:
            if name not in self.independent_vars:
                warnings.warn(f"The keyword argument {name} does not " +
                              "match any arguments of the model function. " +
                              "It will be ignored.", UserWarning)

        # If any parameter is not initialized raise a more helpful error.
        missing_param = any(p not in params.keys() for p in self.param_names)
        blank_param = any((p.value is None and p.expr is None)
                          for p in params.values())
        if missing_param or blank_param:
            msg = ('Assign each parameter an initial value by passing '
                   'Parameters or keyword arguments to fit.\n')
            missing = [p for p in self.param_names if p not in params.keys()]
            blank = [name for name, p in params.items()
                     if p.value is None and p.expr is None]
            msg += f'Missing parameters: {str(missing)}\n'
            msg += f'Non initialized parameters: {str(blank)}'
            raise ValueError(msg)

        # Handle null/missing values.
        if nan_policy is not None:
            self.nan_policy = nan_policy

        mask = None
        if self.nan_policy == 'omit':
            mask = ~isnull(data)
            if mask is not None:
                data = data[mask]
            if weights is not None:
                weights = _align(weights, mask, data)

        # If independent_vars and data are alignable (pandas), align them,
        # and apply the mask from above if there is one.
        for var in self.independent_vars:
            if not np.isscalar(kwargs[var]):
                kwargs[var] = _align(kwargs[var], mask, data)

        if coerce_farray:
            # coerce data and independent variable(s) that are 'array-like' (list,
            # tuples, pandas Series) to float64/complex128.
            data = coerce_arraylike(data)
            for var in self.independent_vars:
                kwargs[var] = coerce_arraylike(kwargs[var])

        if fit_kws is None:
            fit_kws = {}

        output = EMGModelResult(self, params, data=data, weights=weights, 
                                fitted_peaks=fitted_peaks, method=method, 
                                fcn_kws=kwargs, iter_cb=iter_cb, 
                                scale_covar=scale_covar, 
                                nan_policy=self.nan_policy, 
                                calc_covar=calc_covar, 
                                max_nfev=max_nfev, **fit_kws)
        output.fit(data=data, weights=weights) 
        output.components = self.components
        return output
    

    def __add__(self, other):
        """+"""
        return CompositeEMGModel(self, other, operator.add)

    def __sub__(self, other):
        """-"""
        return CompositeEMGModel(self, other, operator.sub)

    def __mul__(self, other):
        """*"""
        return CompositeEMGModel(self, other, operator.mul)

    def __truediv__(self, other):
        """/"""
        return CompositeEMGModel(self, other, operator.truediv)
    


class CompositeEMGModel(EMGModel):
    """Combine two EMGModels (`left` and `right`) with binary operator (`op`).

    Normally, one does not have to explicitly create a `CompositeModel`,
    but can use normal Python operators ``+``, ``-``, ``*``, and ``/`` to
    combine components as in::

    >>> mod = EMGModel(fcn1) + EMGModel(fcn2) * EMGModel(fcn3)

    Class modified from :mod:`lmfit.model.CompositeModel` class.

    """

    _known_ops = {operator.add: '+', operator.sub: '-',
                  operator.mul: '*', operator.truediv: '/'}

    def __init__(self, left, right, op, **kws):
        """Create composite EMG model

        Method modified from :mod:`lmfit.model` module.

        Parameters
        ----------
        left : Model
            Left-hand model.
        right : Model
            Right-hand model.
        op : callable binary operator
            Operator to combine `left` and `right` models.
        **kws : optional
            Additional keywords are passed to `Model` when creating this
            new model.

        Notes
        -----
        The two models can use different independent variables.

        """
        if not isinstance(left, EMGModel):
            raise ValueError(f'CompositeModel: argument {left} is not a Model')
        if not isinstance(right, EMGModel):
            raise ValueError(f'CompositeModel: argument {right} is not a Model')
        if not callable(op):
            raise ValueError(f'CompositeModel: operator {op} is not callable')

        self.left = left
        self.right = right
        self.op = op

        name_collisions = set(left.param_names) & set(right.param_names)
        if len(name_collisions) > 0:
            msg = ''
            for collision in name_collisions:
                msg += (f"\nTwo models have parameters named '{collision}'; "
                        "use distinct names.")
            raise NameError(msg)

        # the unique ``independent_vars`` of the left and right model are
        # combined to ``independent_vars`` of the ``CompositeModel``
        if 'independent_vars' not in kws:
            ivars = self.left.independent_vars + self.right.independent_vars
            kws['independent_vars'] = list(np.unique(ivars))
        if 'nan_policy' not in kws:
            kws['nan_policy'] = self.left.nan_policy
        if 'cost_func' not in kws:
            kws['cost_func'] = self.left.cost_func

        def _tmp(self, *args, **kws):
            pass
        EMGModel.__init__(self, _tmp, **kws)

        for side in (left, right):
            prefix = side.prefix
            for basename, hint in side.param_hints.items():
                self.param_hints[f"{prefix}{basename}"] = hint
    
    def _parse_params(self):
        """Method adopted from :mod:`lmfit.model` module."""
        self._func_haskeywords = (self.left._func_haskeywords or
                                  self.right._func_haskeywords)
        self._func_allargs = (self.left._func_allargs +
                              self.right._func_allargs)
        self.def_vals = deepcopy(self.right.def_vals)
        self.def_vals.update(self.left.def_vals)
        self.opts = deepcopy(self.right.opts)
        self.opts.update(self.left.opts)

    def _reprstring(self, long=False):
        """Method adopted from :mod:`lmfit.model` module."""
        return (f"({self.left._reprstring(long=long)} "
                f"{self._known_ops.get(self.op, self.op)} "
                f"{self.right._reprstring(long=long)})")

    def eval(self, params=None, **kwargs):
        """Evaluate model function for composite model.
        
        Method adopted from :mod:`lmfit.model` module.

        """
        return self.op(self.left.eval(params=params, **kwargs),
                       self.right.eval(params=params, **kwargs))

    def eval_components(self, **kwargs):
        """Return dictionary of name, results for each component.
        
        Method adopted from :mod:`lmfit.model` module.

        """
        out = dict(self.left.eval_components(**kwargs))
        out.update(self.right.eval_components(**kwargs))
        return out

    def post_fit(self, fitresult):
        """function that is called just after fit, can be overloaded by
        subclasses to add non-fitting 'calculated parameters'

        Method adopted from :mod:`lmfit.model` module.

        """
        self.left.post_fit(fitresult)
        self.right.post_fit(fitresult)

    @property
    def param_names(self):
        """Return parameter names for composite model.
        
        Method adopted from :mod:`lmfit.model` module.

        """
        return self.left.param_names + self.right.param_names

    @property
    def components(self):
        """Return components for composite model.
        
        Method adopted from :mod:`lmfit.model` module.

        """
        return self.left.components + self.right.components

    def _get_state(self):
        """Method adopted from :mod:`lmfit.model` module."""
        return (self.left._get_state(),
                self.right._get_state(), self.op.__name__)

    def _set_state(self, state, funcdefs=None):
        """Method adopted from :mod:`lmfit.model` module."""
        return _buildmodel(state, funcdefs=funcdefs)

    def _make_all_args(self, params=None, **kwargs):
        """Generate **all** function arguments for all functions.

        Method adopted from :mod:`lmfit.model` module.

        """
        out = self.right._make_all_args(params=params, **kwargs)
        out.update(self.left._make_all_args(params=params, **kwargs))
        return out



def save_model(model, fname):
    """Save an EMGModel to a file.

    Function adopted from :mod:`lmfit.model` module.

    Parameters
    ----------
    model : :class:`~emgfit.model.EMGModel`
        Model to be saved.
    fname : str
        Name of file for saved Model.

    """
    with open(fname, 'w') as fout:
        model.dump(fout)


def load_model(fname, funcdefs=None): 
    """Load a saved EMGModel from a file.

    Function modified from :mod:`lmfit.model` module.

    Parameters
    ----------
    fname : str
        Name of file containing saved EMGModel.
    funcdefs : dict, optional
        Dictionary of custom function names and definitions.

    Returns
    -------
    Model
        Model object loaded from file.

    """
    m = EMGModel(lambda x: x)
    with open(fname) as fh:
        model = m.load(fh, funcdefs=funcdefs)
    return model


def _buildmodel(state, funcdefs=None):
    """Build EMGModel from saved state.

    Intended for internal use only.

    Function modified from :mod:`lmfit.model` module.

    """
    if len(state) != 3:
        raise ValueError("Cannot restore Model")
    known_funcs = {}
    for fname in lineshapes.functions:
        fcn = getattr(lineshapes, fname, None)
        if callable(fcn):
            known_funcs[fname] = fcn
    if funcdefs is not None:
        known_funcs.update(funcdefs)

    left, right, op = state
    if op is None and right is None:
        (fname, fcndef, cost_func, par_hint_args, vary_baseline, vary_shape, 
         name, prefix, ivars, pnames, 
         phints, nan_policy, opts) = left
        if not callable(fcndef) and fname in known_funcs:
            fcndef = known_funcs[fname]

        if fcndef is None:
            raise ValueError("Cannot restore Model: model function not found")

        if fname == '_eval' and isinstance(fcndef, str):
            raise NotImplementedError
        else:
            model = EMGModel(fcndef, cost_func=cost_func, 
                             par_hint_args=par_hint_args, 
                             vary_baseline=bool(vary_baseline), 
                             vary_shape=bool(vary_shape), 
                             name=name, prefix=prefix,
                             independent_vars=ivars, param_names=pnames,
                             nan_policy=nan_policy, **opts)

        for name, hint in phints.items():
            model.set_param_hint(name, **hint)
        return model
    else:
        lmodel = _buildmodel(left, funcdefs=funcdefs)
        rmodel = _buildmodel(right, funcdefs=funcdefs)
        return CompositeEMGModel(lmodel, rmodel, getattr(operator, op))
    

def save_modelresult(modelresult, fname):
    """Save an EMGModelResult to a file.

    Function adopted from :mod:`lmfit.model` module.

    Parameters
    ----------
    modelresult : :class:`~emgfit.model.EMGModelResult`
        EMGModelResult to be saved.
    fname : str
        Name of file for saved EMGModelResult.

    """
    with open(fname, 'w') as fout:
        modelresult.dump(fout)


def load_modelresult(fname, funcdefs=None):
    """Load a saved EMGModelResult from a file.

    Function modified from :mod:`lmfit.model` module.

    Parameters
    ----------
    fname : str
        Name of file containing saved EMGModelResult.
    funcdefs : dict, optional
        Dictionary of custom function names and definitions.

    Returns
    -------
    :class:`~emgfit.model.EMGModelResult`
        EMGModelResult object loaded from file.

    """
    params = Parameters()
    modres = EMGModelResult(EMGModel(lambda x: x), params)
    with open(fname) as fh:
        mresult = modres.load(fh, funcdefs=funcdefs)
    return mresult



################################################################################
###### Define emgfit EMGModelResult class
class EMGModelResult(lmfit.model.ModelResult):
    """Result from an EMGModel fit."""

    def __init__(self, model, params, data=None, weights=None, 
                 fitted_peaks=None, method='least_squares', fcn_args=None, 
                 fcn_kws=None, iter_cb=None, scale_covar=False, 
                 nan_policy='propagate', calc_covar=False, max_nfev=None, 
                 **fit_kws):
        """
        Parameters
        ----------
        model : Model
            Model to use.
        params : Parameters
            Parameters with initial values for model.
        data : array_like
            Ordinate values of data to be modeled.
        weights : array_like, optional
            Weights to multiply ``(data-model)`` for default fit residual.
        fitted_peaks : list of :class:`emgfit.spectrum.spectrum.peak`
            List of fitted peak objects.
        method : str, optional
            Name of minimization method to use (default is `'least_squares'`).
        fcn_args : sequence, optional
            Positional arguments to send to model function.
        fcn_kws : dict, optional
            Keyword arguments to send to model function.
        iter_cb : callable, optional
            Function to call on each iteration of fit.
        scale_covar : bool, optional
            Whether to scale covariance matrix for uncertainty evaluation.
        nan_policy : {'raise', 'propagate', 'omit'}, optional
            What to do when encountering NaNs when fitting Model.
        calc_covar : bool, optional
            Whether to calculate the covariance matrix (default is True)
            for solvers other than `'leastsq'` and `'least_squares'`.
            Requires the ``numdifftools`` package to be installed.
        max_nfev : int or None, optional
            Maximum number of function evaluations (default is None). The
            default value depends on the fitting method.
        **fit_kws : optional
            Keyword arguments to send to minimization routine.

        """
        if fcn_kws is None:
            fcn_kws = {}
        self.fitted_peaks = fitted_peaks
        self.fit_kws = fit_kws
        super().__init__(model, params, data=data, weights=weights, method=method, 
                         fcn_args=fcn_args, fcn_kws=fcn_kws, iter_cb=iter_cb, 
                         scale_covar=scale_covar, nan_policy=nan_policy, 
                         calc_covar=calc_covar, max_nfev=max_nfev, **fit_kws)
    
    @property 
    def x(self):
        """Ordinate values of fitted data"""
        if self.userkws in [None, {}]:
            x = None
        else:
            x = self.userkws[self.model.independent_vars[0]]
        return x

    @property
    def y(self):
        """Abscissa values of fitted data"""   
        return self.data 
    
    @property 
    def y_err(self):
        """Uncertainties of abscissa values of fitted data"""   
        if self.cost_func == "chi-square":
            # Calculate final weights for plotting
            y_m = self.best_fit
            Pearson_weights = 1./sqrt(y_m + EPS)
            y_err = 1./Pearson_weights
        else:
            y_err = 1./self.weights
        return y_err
    
    @property
    def x_fit_range(self):
        """Range of fitted ordinate values"""   
        return  max(self.x) - min(self.x)
    
    @property
    def x_fit_cen(self):
        """Centre of fitted ordinate range""" 
        return min(self.x) + 0.5*self.x_fit_range
    
    @property 
    def fit_model(self):
        """Name of hyper-EMG model used in fit""" 
        if isinstance(self.model, CompositeEMGModel): 
            fit_model = self.model.components[-1].func.__name__
        else:
            fit_model = self.model.func.__name__
        return fit_model
    
    @property 
    def vary_baseline(self):
        """Whether the amplitude of the uniform background was varied in the fit"""
        return self.model.vary_baseline

    @property 
    def vary_shape(self):
        """Whether the peak shape (and scale) parameters were varied in the fit"""
        return self.model.vary_shape
    
    @property 
    def par_hint_args(self):
        """User-specified parameter hints used in the fit"""
        return self.model.par_hint_args
    
    @property 
    def cost_func(self):
        """Name of the cost function used in the fit"""
        return self.model.cost_func
   

    def fit(self, data=None, fitted_peaks=None, params=None, weights=None, 
            method=None, nan_policy=None, **kwargs):
        """Re-perform fit for a Model, given data and params.  

        Method modified from :mod:`lmfit.model` module.

        Parameters
        ----------
        data : array_like, optional 
            Data to be modeled.
        fitted_peaks : list of :class:`emgfit.spectrum.spectrum.peak`
            List of peak objects to be fitted.
        params : Parameters, optional
            Parameters with initial values for model.
        weights : array_like, optional
            Weights to multiply ``(data-model)`` for fit residual - only 
            relevant when using the default lmfit cost function. 
        method : str, optional
            Name of minimization method to use (default is `'least_squares'`).
        nan_policy : {'raise', 'propagate', 'omit'}, optional
            What to do when encountering NaNs when fitting Model.
        **kwargs : optional
            Keyword arguments to send to minimization routine.

        """
        if data is not None:
            self.data = data
        if fitted_peaks is not None:
            self.fitted_peaks = fitted_peaks
        if params is not None:
            self.init_params = params
        if weights is not None:
            self.weights = weights
        if method is not None:
            self.method = method
        if nan_policy is not None:
            self.nan_policy = nan_policy

        self.ci_out = None
        self.userargs = (self.data, self.weights)
        self.userkws.update(kwargs)
        self.init_fit = self.model.eval(params=self.params, **self.userkws)
        _ret = self.minimize(method=self.method)
        self.model.post_fit(_ret)
        _ret.params.create_uvars(covar=_ret.covar)

        for attr in dir(_ret):
            if not attr.startswith('_') and attr != "x":
                try:
                    setattr(self, attr, getattr(_ret, attr))
                except AttributeError:
                    pass

        self.init_values = self.model._make_all_args(self.init_params)
        self.best_values = self.model._make_all_args(_ret.params)
        self.best_fit = self.model.eval(params=_ret.params, **self.userkws)
        if (self.data is not None and len(self.data) > 1
           and isinstance(self.best_fit, np.ndarray)
           and len(self.best_fit) > 1):
            dat = coerce_arraylike(self.data)
            resid = ((dat - self.best_fit)**2).sum()
            sstot = ((dat - dat.mean())**2).sum()
            self.rsquared = 1.0 - resid/max(tiny, sstot)
    
    def __repr__(self):
        """Return representation of EMGModelResult.
        
        Method modified from :mod:`lmfit.model` module.

        """
        return f"<emgfit.EMGModelResult: {self.model.name}>"
    
    def loads(self, s, funcdefs=None, **kws):
        """Load ModelResult from a JSON string.

        Parameters
        ----------
        s : str
            String representation of ModelResult, as from `dumps`.
        funcdefs : dict, optional
            Dictionary of custom function names and definitions.
        **kws : optional
            Keyword arguments that are passed to `json.loads`.

        Returns
        -------
        ModelResult
            ModelResult instance from JSON string representation.

        See Also
        --------
        load, dumps, json.dumps

        """
        modres = json.loads(s, **kws)
        if 'modelresult' not in modres['__class__'].lower():
            raise AttributeError('ModelResult.loads() needs saved ModelResult')

        modres = decode4js(modres)
        if 'model' not in modres or 'params' not in modres:
            raise AttributeError('ModelResult.loads() needs valid ModelResult')

        # model
        self.model = _buildmodel(decode4js(modres['model']), funcdefs=funcdefs)

        if funcdefs:
            # Remove model function so as not pass it into the _asteval.symtable
            funcdefs.pop(self.model.func.__name__, None)

        # how params are saved was changed with version 2:
        modres_vers = modres.get('__version__', '1')
        if modres_vers == '1':
            for target in ('params', 'init_params'):
                state = {'unique_symbols': modres['unique_symbols'], 'params': []}
                for parstate in modres['params']:
                    _par = Parameter(name='')
                    _par.__setstate__(parstate)
                    state['params'].append(_par)
                _params = Parameters(usersyms=funcdefs)
                _params.__setstate__(state)
                setattr(self, target, _params)

        elif modres_vers == '2':
            for target in ('params', 'init_params'):
                _pars = Parameters()
                _pars.loads(modres[target])
                if funcdefs:
                    for key, val in funcdefs.items():
                        _pars._asteval.symtable[key] = val
                setattr(self, target, _pars)

        for attr in ('aborted', 'aic', 'best_fit', 'best_values', 'bic',
                     'chisqr', 'ci_out', 'col_deriv', 'covar', 'data',
                     'errorbars', 'fjac', 'flatchain', 'ier', 'init_fit',
                     'init_values', 'kws', 'lmdif_message', 'message',
                     'method', 'nan_policy', 'ndata', 'nfev', 'nfree',
                     'nvarys', 'redchi', 'residual', 'rsquared', 'scale_covar',
                     'calc_covar', 'success', 'userargs', 'userkws',
                     'var_names', 'weights', 'user_options', 'fitted_peaks', 
                     'fit_kws'):
            setattr(self, attr, decode4js(modres.get(attr, None)))

        self.best_fit = self.model.eval(self.params, **self.userkws)
        if len(self.userargs) == 2:
            self.data = self.userargs[0]
            self.weights = self.userargs[1]

        for parname, val in self.init_values.items():
            par = self.init_params.get(parname, None)
            if par is not None:
                par.correl = par.stderr = None
                par.value = par.init_value = self.init_values[parname]

        self.init_fit = self.model.eval(self.init_params, **self.userkws)
        self.result = lmfit.minimizer.MinimizerResult()
        self.result.params = self.params

        if self.errorbars and self.covar is not None:
            self.uvars = self.result.params.create_uvars(covar=self.covar)

        self.init_vals = list(self.init_values.items())
        return self

    def load(self, fp, funcdefs=None, **kws):
        """Load JSON representation of ModelResult from a file-like object.

        Parameters
        ----------
        fp : file-like object
            An open and `.read()`-supporting file-like object.
        funcdefs : dict, optional
            Dictionary of function definitions to use to construct Model.
        **kws : optional
            Keyword arguments that are passed to `loads`.

        Returns
        -------
        ModelResult
            ModelResult created from `fp`.

        See Also
        --------
        dump, loads, json.load

        """
        return self.loads(fp.read(), funcdefs=funcdefs, **kws)

    


import pytest
import emgfit as emg
import numpy as np

class Test_get_AME_values:
    """Tests of emgfit literature values against reference values

    Electron binding energies and the uncertainty of m_e are neglected!

    """
    decimals = 8 # number of decimals up to which absolute agreement is demanded
    atol = 10**(-decimals)
    rtol = 0 # relative tolerance up to which agreement is demanded
    m_e_ref = 0.000548579909065 # CODATA value from physics.nist.gov
    u_to_keV_ref = 931494.10242  # CODATA value from physics.nist.gov
    lit_src = 'AME2020'

    def _isclose(self, a ,b, atol=None, rtol=None):
        """Custom version of np.isclose. """
        abs_tol = atol or self.atol
        rel_tol = rtol or self.rtol
        return np.isclose(a, b, atol=abs_tol, rtol=rel_tol)


    def _check_lit_values(self, species, m_ref, m_ref_error, extrapol_ref, A_ref,
                          z_ref=1, Ex=0.0, Ex_error=0.0):
        """Helper method for checking agreement with reference values """
        from emgfit.ame_funcs import get_AME_values, get_charge_state
        m_AME, m_AME_error, extrapol, A  = get_AME_values(species,
                                                          Ex=Ex,
                                                          Ex_error=Ex_error,
                                                          src=self.lit_src)
        z = get_charge_state(species)

        msg0 = "`m_AME` of {} deviates from reference".format(species)
        assert self._isclose(m_AME, m_ref), msg0
        msg1 = "`m_AME_error` of {} deviates from reference".format(species)
        assert self._isclose(m_AME_error, m_ref_error), msg1
        msg2 = "wrong extrapol. flag for {}".format(species)
        assert extrapol == extrapol_ref, msg2
        msg3 = "atomic mass number of {} deviates from reference".format(species)
        assert A == A_ref, msg3
        msg4 = "result of get_charge_state() deviates from reference"
        assert z == z_ref, msg4


    def check_u_to_keV_conversion(self):
        """Check the u_to_keV factor from Scipy against value from
        """
        msg = "u_to_keV factor deviates from CODATA reference value"
        assert self._isclose(emg.u_to_keV, self.u_to_keV_ref,atol=5,rtol=0), msg


    def test_extrapol_AME_value(self):
        """Check calculation of ionic mass of extrapolated superheavy species
        """
        species = '1Bh264:-1e'
        # atomic masses and m_e from AME2020 [u]:
        m_ref = 264.124486 - self.m_e_ref
        m_ref_error =  190e-06
        extrapol_ref = True
        A_ref = 264
        z_ref = 1

        self._check_lit_values(species, m_ref, m_ref_error, extrapol_ref, A_ref,
                               z_ref=z_ref, Ex=0.0, Ex_error=0.0)


    def test_molecular_AME_values(self):
        """Check calculation of ionic mass of H2O (incl. tentative ID flag)"""
        species = '2H1:1O16?:-1e'
        # atomic masses and m_e from AME2020 [u]:
        m_H1_ref = 1.007825031898
        m_H1_ref_error = 0.000014e-06
        m_O16_ref = 15.99491461926
        m_O16_ref_error = 0.00032e-06
        m_ref = 2*m_H1_ref + m_O16_ref - self.m_e_ref
        m_ref_error = np.sqrt((2*m_H1_ref_error)**2 + m_O16_ref_error**2)
        extrapol_ref = False
        A_ref = 18
        z_ref = 1

        self._check_lit_values(species, m_ref, m_ref_error, extrapol_ref, A_ref,
                               z_ref=z_ref, Ex=0.0, Ex_error=0.0)


    def test_isomer_AME_values(self):
        """Check calculation of ionic mass of second isomer of In-127 """
        species = '1In127m1:-1e'
        # g.s. mass and m_e from AME2020 [u], Ex from ENSDF [keV]:
        Ex = 1863
        Ex_error = 58
        m_In127_ref = 126.917466040
        m_In127_ref_error = 10.736e-06
        m_ref =  m_In127_ref + Ex/self.u_to_keV_ref - self.m_e_ref
        m_ref_error = np.sqrt(m_In127_ref_error**2 +
                              (Ex_error/self.u_to_keV_ref)**2)
        extrapol_ref = False
        A_ref = 127
        z_ref = 1

        self._check_lit_values(species, m_ref, m_ref_error, extrapol_ref, A_ref,
                               z_ref=z_ref, Ex=Ex, Ex_error=Ex_error)


    def test_molecular_isomer_AME_values(self):
        """Check calculation of ionic mass of second isomer of Sr85m:F19:-1e """
        species = 'Sr85m:F19:-1e'
        # g.s. mass and m_e from AME2020 [u], Ex from ENSDF [keV]:
        Ex = 238.79
        Ex_error = 0.05
        m_Sr85_ref = 84.912932041
        m_Sr85_ref_error = 3.020e-06
        m_F19_ref = 18.99840316207
        m_F19_ref_error = 0.00088e-06
        m_ref = m_Sr85_ref + Ex/self.u_to_keV_ref + m_F19_ref - self.m_e_ref
        m_ref_error = np.sqrt(m_Sr85_ref_error**2 +
                              (Ex_error/self.u_to_keV_ref)**2 +
                              m_F19_ref_error**2)
        extrapol_ref = False
        A_ref = 104
        z_ref = 1

        self._check_lit_values(species, m_ref, m_ref_error, extrapol_ref, A_ref,
                               z_ref=z_ref, Ex=Ex, Ex_error=Ex_error)


    def test_doubly_charged_AME_values(self):
        """Check calculation of ionic mass of doubly charged Sn-116 """
        species = 'Sn116:-2e'
        # atomic mass and m_e from AME2020 [u]:
        m_ref = 115.901742825 - 2*self.m_e_ref
        m_ref_error = 0.103e-06
        extrapol_ref = False
        A_ref = 116
        z_ref = 2

        self._check_lit_values(species, m_ref, m_ref_error, extrapol_ref, A_ref,
                               z_ref=z_ref)


    def test_doubly_charged_molecular_isomer_AME_values(self):
        """Check calculation of ionic mass of first isomer of 2Y89m:O16:-2e """
        species = '2Y89m:1O16:-2e'
        # g.s. mass and m_e from AME2020 [u], Ex from ENSDF [keV]:
        Ex = 908.97
        Ex_error = 0.03
        m_Y89_ref = 88.905838156
        m_Y89_ref_error = 0.363e-06
        m_O16_ref = 15.99491461926
        m_O16_ref_error = 0.00032e-06
        m_ref = 2*(m_Y89_ref + Ex/self.u_to_keV_ref) + m_O16_ref -2*self.m_e_ref
        m_ref_error = np.sqrt((2*m_Y89_ref_error)**2 +
                              (2*Ex_error/self.u_to_keV_ref)**2 +
                              (m_O16_ref_error)**2)
        extrapol_ref = False
        A_ref = 194
        z_ref = 2

        self._check_lit_values(species, m_ref, m_ref_error, extrapol_ref, A_ref,
                               z_ref=z_ref, Ex=Ex, Ex_error=Ex_error)


    def test_get_El_from_Z(self):
        """Check grabbing of element symbol from proton number """
        proton_numbers = [0,1,6,67,118]
        El_ref = ['n','H','C','Ho','Og']
        from emgfit.ame_funcs import get_El_from_Z
        for i, Z in enumerate(proton_numbers):
            El = get_El_from_Z(Z)
            msg = str("result of get_El_from_Z({}) deviates from "
                      "reference").format(Z)
            assert El==El_ref[i], msg
